/*
 * PI1 image loader
 *
 * Copyright (C) 2004 - 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */

#include <abydos-plugin.h>
#include <endian.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>                     /* malloc() */
#include <string.h>                     /* strcasecmp() */

#include "debug.h"

#define ENABLE_PC1

#define PI1_ANIM_LEFT 0
#define PI1_ANIM_OFF 1
#define PI1_ANIM_RIGHT 2

typedef struct{
    uint16_t left_color[4];
    uint16_t right_color[4];
    uint16_t direction[4];
    uint16_t delay[4];
}degas_animation_t;

typedef struct{
    uint16_t resolution;
    uint16_t palette[16];
    uint8_t packed[];
}pc1_t;

typedef struct{
    uint16_t resolution;
    uint16_t palette[16];
    uint16_t image[16000];
    degas_animation_t animation;
}pi1_t;

typedef struct{
    int animmask;
    double duration;
}frame_t;

typedef cairo_surface_t*(*decode_image_func_t)(abydos_plugin_handle_t *, uint16_t *atari_image, uint32_t *cairo_palette);

struct _abydos_plugin_handle_t{
#if 0
    enum{
        S_RESOLUTION = 0,
        S_PALETTE,
        S_IMAGE,
        S_DONE
    }state;
    int pos;
    int subpos;
    int xpad;
    GdkPixbufModuleSizeFunc size_func;
    GdkPixbufModulePreparedFunc prepared_func;
    GdkPixbufModuleUpdatedFunc updated_func;
    void userdata;
#endif
    abydos_plugin_info_t *info;
    enum{
        T_PI1,
        T_PC1
    }type;
    struct{
        int left_color;
        int right_color;
        int steps;
        int direction;
        int delay;
        double frame_duration;
        double total_duration;
    }anim[4];
    frame_t *frame;
    int resolution;
    decode_image_func_t decode_image;
    uint16_t atari_palette[16];
    uint16_t atari_image[16000];
    uint32_t cairo_palette[16];
    cairo_surface_t *cached_surface;
};


#if 0
static gpointer begin_load(
    GdkPixbufModuleSizeFunc size_func,
    GdkPixbufModulePreparedFunc prepared_func,
    GdkPixbufModuleUpdatedFunc updated_func,
    gpointer userdata,
#ifdef TEST_MATCH_INFO
    GError** error,
    GdkPixbufModuleMatchInfo* match_info
#else
    GError** error
#endif
){
    Context* context;

    DEBUG_p("begin");
#ifdef TEST_MATCH_INFO
    DEBUG_p("I was chosen because:");
    if(match_info->mime_type)
        DEBUG_p("mime-type is: %s",match_info->mime_type);
    if(match_info->pattern_index>=0)
        DEBUG_p("pattern %d matched",match_info->pattern_index);
    if(match_info->extension_index>=0)
        DEBUG_p("extension %d matched",match_info->extension_index);
#endif

    context = g_new(Context,1);
    context->size_func = size_func;
    context->prepared_func = prepared_func;
    context->updated_func = updated_func;
    context->userdata = userdata;
    context->state = 0;
    context->pos = 0;
    return (gpointer)context;
}
#endif


#if 0
static gboolean
stop_load(gpointer handle,GError** error)
{
    DEBUG_p("begin");
    g_free(handle);
    return TRUE;
}
#endif


#if 0
static gboolean
load_increment(gpointer handle,
               const guchar* data,
               guint data_len,
               GError** error)
{
    Context* context;
    int data_pos;
    int bytes;

    DEBUG_p("begin");
    context = (Context*)handle;
    data_pos = 0;

    while(data_pos < data_len){
        switch(context->state){
        case S_RESOLUTION:
            bytes = MIN(data_len-data_pos,2-context->pos);
            if(bytes){
                memcpy(context->pos + (char*)&context->resolution,data+data_pos,bytes);
                context->pos += bytes;
            }else{
                context->resolution = ntohs(context->resolution);
                DEBUG_p("resolution = %i",context->resolution);

                /* FIXME!! create pixbuf depending on the resolution */
                context->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,FALSE,8,320,200);
                DEBUG_p("about to call prepared");
                context->prepared_func(context->pixbuf,NULL,context->userdata);
                context->xpad = gdk_pixbuf_get_rowstride(context->pixbuf) - 320*3;
                DEBUG_p("xpad = %i",context->xpad);

                context->pos = 0;
                context->state = S_PALETTE;
            }
            break;
        case S_PALETTE:
            bytes = MIN(data_len-data_pos,32-context->pos);
            if(bytes){
                memcpy(context->pos + (char*)context->atari_palette,data+data_pos,bytes);
                context->pos += bytes;
            }else{
                DEBUG_p("palette sucessfully read");
                {
                    int i;
                    for(i=0;i<16;++i){
                        context->real_palette[i*3+0] = ((context->atari_palette[i]>>0x0)&0x07)*9*4; /* red */
                        context->real_palette[i*3+1] = ((context->atari_palette[i]>>0xc)&0x07)*9*4; /* green */
                        context->real_palette[i*3+2] = ((context->atari_palette[i]>>0x8)&0x07)*9*4; /* blue */
                    }
                }
                context->pos = 0;
                context->state = S_IMAGE;
            }
            break;
        case S_IMAGE:
            bytes = MIN(data_len-data_pos,32000-context->pos);
            /* DEBUG_p("placing %i bytes at %i",bytes,context->pos); */
            if(bytes){
                int first_line;
                int last_line;
                int line_count;
                int first_pos;
                int last_pos;
                int pixel;
                int plane1;
                int plane2;
                int plane3;
                int plane4;
                int i;
                int j;
                guint8* dst;

                memcpy(context->pos + (char*)context->atari_image,data+data_pos,bytes);
                first_pos = context->pos & ~7;
                context->pos += bytes;
                last_pos = context->pos & ~7;
                /* FIXME!! update pixmap dynamically */

                first_line = first_pos/160;
                last_line = (last_pos-1)/160;

                DEBUG_p("updating pixels from %i to %i",first_pos,last_pos);
                dst = gdk_pixbuf_get_pixels(context->pixbuf)+first_pos*2*3;
                for(i=first_pos>>1;i<last_pos>>1;i+=4){
                    plane1 = ntohs(context->atari_image[i+0]);
                    plane2 = ntohs(context->atari_image[i+1]);
                    plane3 = ntohs(context->atari_image[i+2]);
                    plane4 = ntohs(context->atari_image[i+3]);
                    for(j=15;j>=0;--j){
                        pixel = (plane1 & (1<<j) ? 0x01 : 0) |
                                (plane2 & (1<<j) ? 0x02 : 0) |
                                (plane3 & (1<<j) ? 0x04 : 0) |
                                (plane4 & (1<<j) ? 0x08 : 0);
                        memcpy(dst,&context->real_palette[pixel*3],3);
                        dst+=3;
                    }
                }
                line_count = last_line - first_line + 1;
                DEBUG_p("line_count = %i",line_count);
                context->updated_func(context->pixbuf,
                                      0,first_line,
                                      320,line_count,
                                      context->userdata);

            }else{
                context->pos = 0;
                context->state = S_DONE;
            }
            break;
        case S_DONE:
            DEBUG_p("ignoring %i trailing bytes",data_len-data_pos);
            /* bytes = data_len-data_pos; */
            return TRUE;
        }
        data_pos += bytes;
    }
    return TRUE;
}
#endif



static const char **
_pi1_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_PC1
        "image/x-pc1",
        "image/x-pc2",
        "image/x-pc3",
#endif
        "image/x-pi1",
        "image/x-pi2",
        "image/x-pi3",
        NULL
    };
    return mime_types;
}


#ifdef ENABLE_PC1
static int
_packbits_unpack(uint8_t *src, int len, uint16_t *atari_image, int resolution)
{
    uint8_t linebuffer[160];
    int x;
    int y;
    for(y=0;y<200;++y){
        for(x=0;x<160;){
            int b;
            if(len<2){
                DEBUG_p("out of data");
                return -1;
            }
            b = *src++;
            --len;
            if(b>=128){
                b = 257-b;
                if(x+b>160){
                    DEBUG_p("invalid runlength");
                    return -1;
                }
                memset(linebuffer+x,*src,b);
                ++src;
            }else{
                ++b;
                if(x+b>160){
                    DEBUG_p("invalid data");
                    return -1;
                }else if(b>len){
                    DEBUG_p("out of data");
                    return -1;
                }
                memcpy(linebuffer+x,src,b);
                src += b;
            }
            x += b;
        }
        switch(resolution){
        case 0:
            for(x=0;x<20;++x){
                atari_image[y*80+x*4+0] = ((uint16_t*)linebuffer)[x+0];
                atari_image[y*80+x*4+1] = ((uint16_t*)linebuffer)[x+20];
                atari_image[y*80+x*4+2] = ((uint16_t*)linebuffer)[x+40];
                atari_image[y*80+x*4+3] = ((uint16_t*)linebuffer)[x+60];
            }
            break;
        case 1:
            for(x=0;x<40;++x){
                atari_image[y*80+x*2+0] = ((uint16_t*)linebuffer)[x+0];
                atari_image[y*80+x*2+1] = ((uint16_t*)linebuffer)[x+40];
            }
            break;
        case 2:
            memcpy(atari_image+y*80,linebuffer,160);
            break;
        }
    }
    return len;
}
#endif


static void
_atari_decode_palette(abydos_plugin_handle_t *h, uint16_t *atari_palette, uint32_t *cairo_palette)
{
    int i;
    for(i=0;i<16;++i){
        uint16_t col = be16toh(atari_palette[i]);
        int r = (col>>8)&0x07;
        int g = (col>>4)&0x07;
        int b = (col>>0)&0x07;
        r = (r<<5) + (r<<2) + (r>>1);
        g = (g<<5) + (g<<2) + (g>>1);
        b = (b<<5) + (b<<2) + (b>>1);
        cairo_palette[i] = (r<<16) + (g<<8) + b;
    }
}

static cairo_surface_t *
_atari_decode_image_0(abydos_plugin_handle_t *h, uint16_t *atari_image, uint32_t *cairo_palette)
{
    cairo_surface_t *surface;
    uint32_t* dst;
    int rowskip;
    int x;
    int y;
    int i;
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,320,200);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-320;
    for(y=0;y<200;++y){
        for(x=0;x<20;++x){
            int plane1 = be16toh(atari_image[y*80+x*4+0]);
            int plane2 = be16toh(atari_image[y*80+x*4+1]);
            int plane3 = be16toh(atari_image[y*80+x*4+2]);
            int plane4 = be16toh(atari_image[y*80+x*4+3]);
            for(i=15;i>=0;--i){
                int pixel = (plane1 & (1<<i) ? 0x01 : 0) |
                            (plane2 & (1<<i) ? 0x02 : 0) |
                            (plane3 & (1<<i) ? 0x04 : 0) |
                            (plane4 & (1<<i) ? 0x08 : 0);
                *dst++ = cairo_palette[pixel];
            }
        }
        dst += rowskip;
    }
    return surface;
}


static cairo_surface_t *
_atari_decode_image_1(abydos_plugin_handle_t *h, uint16_t *atari_image, uint32_t *cairo_palette)
{
    cairo_surface_t *surface;
    uint32_t* dst;
    int rowskip;
    int x;
    int y;
    int i;
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,640,200);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-640;
    for(y=0;y<200;++y){
        for(x=0;x<40;++x){
            int plane1 = be16toh(atari_image[y*80+x*2+0]);
            int plane2 = be16toh(atari_image[y*80+x*2+1]);
            for(i=15;i>=0;--i){
                int pixel = (plane1 & (1<<i) ? 0x01 : 0) |
                            (plane2 & (1<<i) ? 0x02 : 0);
                *dst++ = cairo_palette[pixel];
            }
        }
        dst += rowskip;
    }
    return surface;
}


static cairo_surface_t *
_atari_decode_image_2(abydos_plugin_handle_t *h, uint16_t *atari_image, uint32_t *cairo_palette)
{
    cairo_surface_t *surface;
    uint32_t* dst;
    int rowskip;
    int x;
    int y;
    int i;
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,640,400);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-640;
    for(y=0;y<400;++y){
        for(x=0;x<40;++x){
            int plane1 = be16toh(atari_image[y*40+x]);
            for(i=15;i>=0;--i)
                *dst++ = plane1 & (1<<i) ? 0 : 0xffffff;
        }
        dst += rowskip;
    }
    return surface;
}


static void
_setup_resolution(abydos_plugin_handle_t *h)
{
    switch(h->resolution){
    case 0:
    case 3:
        h->info->width = 320;
        h->info->height = 200;
        h->info->pixel_ratio = 5.0/6;
        h->decode_image = _atari_decode_image_0;
        break;
    case 1:
        h->info->width = 640;
        h->info->height = 200;
        h->info->pixel_ratio = 5.0/12;
        h->decode_image = _atari_decode_image_1;
        break;
    case 2:
        h->info->width = 640;
        h->info->height = 400;
        h->info->pixel_ratio = 5.0/6;
        h->decode_image = _atari_decode_image_2;
        break;
    }
}


static void
_setup_animation(abydos_plugin_handle_t *h, degas_animation_t *a)
{
    int i;
    for(i=0;i<4;++i){
        h->anim[i].left_color = be16toh(a->left_color[i]);
        h->anim[i].right_color = be16toh(a->right_color[i]);
        h->anim[i].steps = 1 + h->anim[i].right_color - h->anim[i].left_color;
        h->anim[i].direction = be16toh(a->direction[i]);
        //DEBUG_p("delay value is %d",(int)ntohs(a->delay[i]));
        h->anim[i].delay = 128 - be16toh(a->delay[i]);
        //DEBUG_p("delay is %d",h->anim[i].delay);
        h->anim[i].frame_duration = (double)h->anim[i].delay / 60;
        //DEBUG_p("duration is %f seconds",h->anim[i].frame_duration);
        h->anim[i].total_duration = h->anim[i].frame_duration * h->anim[i].steps;
    }
    if(h->anim[0].steps > 1 && h->anim[0].direction != PI1_ANIM_OFF){
        h->info->frame_count = h->anim[0].steps;
        h->frame = (frame_t*)malloc(h->info->frame_count * sizeof(frame_t));
        for(i=0;i<h->info->frame_count;++i){
            h->frame[i].animmask = i>0 ? (1<<0) : 0;
            h->frame[i].duration = h->anim[0].frame_duration;
        }
    }
}


static void
_animate_palette(abydos_plugin_handle_t *h, double time, uint32_t *palette)
{
    int i;
    for(i=0;time>=0 && i<h->info->frame_count;++i){
        int a;
        uint32_t color;
        for(a=0;a<4;++a)
            if(h->frame[i].animmask & (1<<a)){
                if(PI1_ANIM_LEFT == h->anim[0].direction){
                    color = palette[h->anim[0].left_color];
                    memmove(&palette[h->anim[0].left_color],&palette[h->anim[0].left_color+1],(h->anim[0].steps-1)*sizeof(uint32_t));
                    palette[h->anim[0].right_color] = color;
                }else if(PI1_ANIM_RIGHT == h->anim[0].direction){
                    color = palette[h->anim[0].right_color];
                    memmove(&palette[h->anim[0].left_color+1],&palette[h->anim[0].left_color],(h->anim[0].steps-1)*sizeof(uint32_t));
                    palette[h->anim[0].left_color] = color;
                }
            }
        time -= h->frame[i].duration;
    }
}


static abydos_plugin_handle_t *
_pi1_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));;
#ifdef ENABLE_PC1
    if(!strcasecmp(mime_type,"image/x-pc1")||
       !strcasecmp(mime_type,"image/x-pc2")||
       !strcasecmp(mime_type,"image/x-pc3"))
    {
        h->type = T_PC1;
    }else
#endif
    {
        h->type = T_PI1;
    }
    h->info = info;
    h->cached_surface = NULL;
    info->threadsafe = 1;
    return h;
}


static void
_pi1_free(abydos_plugin_handle_t *h)
{
    if(h->cached_surface)
        cairo_surface_destroy(h->cached_surface);
    free(h);
}



static void
_pi1_get_duration(abydos_plugin_handle_t *h, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}



static int
_pi1_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{

    switch(h->type){
    case T_PI1:
        if(len < offsetof(pi1_t,animation)){
            h->info->error = "data is too short";
            return -1;
        }
        {
            pi1_t *pi1 = (pi1_t*)data;
            h->resolution = be16toh(pi1->resolution)&3;
            _setup_resolution(h);
            if(len >= sizeof(pi1_t)){
                /* animation is included */
                DEBUG_p("animation is included");
                _setup_animation(h, &pi1->animation);
            }
            if(h->info->frame_count > 1){
                memcpy(h->atari_palette,pi1->palette,16*sizeof(uint16_t));
                memcpy(h->atari_image,pi1->image,16000*sizeof(uint16_t));
            }else{
                uint32_t cairo_palette[16];
                _atari_decode_palette(h, pi1->palette, cairo_palette);
                h->cached_surface = h->decode_image(h, pi1->image, cairo_palette);
            }
        }
        break;
#ifdef ENABLE_PC1
    case T_PC1:
        {
            pc1_t *pc1 = (pc1_t*)data;
            int res;
            h->resolution = be16toh(pc1->resolution)&3;
            _setup_resolution(h);
            res = _packbits_unpack(pc1->packed,len-sizeof(pc1_t),h->atari_image,h->resolution);
            if(res<0){
                h->info->error = "invalid data";
                return res;
            }
            if(res >= sizeof(degas_animation_t))
                _setup_animation(h, (degas_animation_t*)(data+len-sizeof(degas_animation_t)));
            if(h->info->frame_count > 1){
                memcpy(h->atari_palette,pc1->palette,16*sizeof(uint16_t));
            }else{
                uint32_t cairo_palette[16];
                _atari_decode_palette(h, pc1->palette, cairo_palette);
                h->cached_surface = h->decode_image(h, h->atari_image, cairo_palette);
            }
        }
        break;
#endif
    }
    return 0;
}


static cairo_surface_t *
_pi1_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    if(h->info->frame_count > 1){
        uint32_t cairo_palette[16];
        _atari_decode_palette(h, h->atari_palette, cairo_palette);
        _animate_palette(h,time,cairo_palette);
        return h->decode_image(h, h->atari_image, cairo_palette);
    }else{
        return cairo_surface_reference(h->cached_surface);
    }
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _pi1_get_mime_types,
        .new = _pi1_new,
        .free = _pi1_free,
        .create_from_data = _pi1_create_from_data,
        .get_image_surface = _pi1_get_image_surface,
        .get_duration = _pi1_get_duration
    };
    DEBUG_p("begin");
    return &p;
}
