/*
 * Copyright (C) 2018 - 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

#include "debug.h"

/*

*** Standard C64 BITMAP Files (HIRES & HIRES-MULTICOLOR)
*** Document revision: 1.3
*** Last updated: March 11, 2004
*** Compiler/Editor: Peter Schepers
*** Contributors/sources: Peter Weighill

  Below is a table to help in the identification of C64  bitmap  files.  If
you require further information about how bitmaps are  displayed  then  you
should refer to the C64 Programmers Reference Manual.

  To automatically identify the type of  file,  you  need  to  compare  the
length of the file and the Load address. Once you have identified the  file
type, you can then use the bitmap, screen and colour offsets to help in the
displaying of the bitmap.

Format for file types:
       Load   - Load address (first two bytes of the file)
       Length - File length in bytes (including load address bytes)
       Bitmap - Bitmap offset (8000 bytes)
       Screen - Screen data offset (1000 bytes)
       Colour - Colour data offset (1000 bytes)
       ScrCol - Screen colour offset (1 byte)

Name (Black/white)    Load  Length Bitmap Screen Colour ScrCol
--------------------- ----- ------ ------ ------ ------ ------
"Art Studio"          $2000 9009   0      8000   -      -
"Doodle"              $5C00 9218   1024   0      -      -
"Image System (hi)"   $4000 9194   0      8192   -      -

Name (Multicolor)     Load  Length Bitmap Screen Colour ScrCol
--------------------- ----- ------ ------ ------ ------ ------
"Artist 64"           $4000 10242  0      8192   9216   10239
"Adv Art Studio"      $2000 10018  0      8000   9016   9001
"Blazing Paddles"     $A000 10242  0      8192   9216   8064
"Koala Painter"       $6000 10003  0      8000   9000   10000
"Image System (mc)"   $3C00 10218  1024   9216   0      9215
"Vidcom 64"           $5800 10050  2048   1024   0      2024
"CDU Paint"           $7EEF 10277  273    8273   9273   10273

*/


typedef struct{
    uint8_t bitmap[8000];
    uint8_t screen[1000];
    uint8_t color[1000];
    uint8_t background;
}vmem_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_koala_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-koa",
        "image/x-koala",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_koala_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h;
    h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->surface = NULL;
    info->threadsafe = 1;
    return h;
}


static void
_koala_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static uint32_t
_get_color(vmem_t *vmem,int x,int y,int c)
{
#if 1
    /* Colodore */
    static const uint32_t color[16] = {
        0x000000, /* black */
        0xffffff, /* white */
        0x813338, /* dark red */
        0x75cec8, /* cyan */
        0x8e3c97, /* purple */
        0x56ac4d, /* dark green */
        0x2e2c9b, /* dark blue */
        0xedf171, /* yellow */
        0x8e5029, /* orange */
        0x553800, /* brown */
        0xc46c71, /* light red */
        0x4a4a4a, /* dark gray */
        0x7b7b7b, /* mid gray */
        0xa9ff9f, /* light green */
        0x706deb, /* light blue */
        0xb2b2b2  /* light gray */
    };
#else
    static const uint32_t color[16] = {
        0x000000, /* black */
        0xffffff, /* white */
        0x880000, /* dark red */
        0xaaffee, /* cyan */
        0xcc44cc, /* purple */
        0x00cc55, /* dark green */
        0x0000aa, /* dark blue */
        0xeeee77, /* yellow */
        0xdd8855, /* orange */
        0x664400, /* brown */
        0xff7777, /* light red */
        0x333333, /* dark gray */
        0x777777, /* mid gray */
        0xaaff66, /* light green */
        0x0088ff, /* light blue */
        0xbbbbbb  /* light gray */
    };
#endif
    switch(c){
    default:
    case 0:return color[vmem->background];
    case 1:return color[vmem->screen[x+40*y]>>4];
    case 2:return color[vmem->screen[x+40*y]&15];
    case 3:return color[vmem->color[x+40*y]&15];
    }
}


static cairo_surface_t *
_c64_decode_image(abydos_plugin_handle_t *h,vmem_t *vmem)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,160,200);
    uint8_t *src;
    uint32_t *dst;
    int stride;
    int x;
    int y;
    int i;
    if(!surface)
        return NULL;
    src = vmem->bitmap;
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    for(y=0;y<25;++y){
        for(x=0;x<40;++x){
            for(i=0;i<8;++i){
                int pix = *src++;
                dst[x*4+0+(y*8+i)*stride] = _get_color(vmem,x,y,(pix>>6) & 3);
                dst[x*4+1+(y*8+i)*stride] = _get_color(vmem,x,y,(pix>>4) & 3);
                dst[x*4+2+(y*8+i)*stride] = _get_color(vmem,x,y,(pix>>2) & 3);
                dst[x*4+3+(y*8+i)*stride] = _get_color(vmem,x,y,(pix>>0) & 3);
            }
        }
    }
    return surface;
}


static int
_koala_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    DEBUG_p("begin (size = %d)",(int)len);
    if(len != 10003)
        return -1;
    h->info->width = 160;
    h->info->height = 200;
    h->info->pixel_ratio = 5.0/3;
    h->surface = _c64_decode_image(h, (vmem_t*)&data[2]);
    return 0;
}


static cairo_surface_t *
_koala_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _koala_get_mime_types,
        .new = _koala_new,
        .free = _koala_free,
        .create_from_data = _koala_create_from_data,
        .get_image_surface = _koala_get_image_surface
    };
    return &p;
}
