/*
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <gd.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

/* #define ENABLE_BMP */
/* #define ENABLE_PNG */
#define ENABLE_WBMP
/* #define ENABLE_WEBP */ /* not working */
#define ENABLE_XBM

/*

Supported formats

* BMP
  GD     MIME-type?
  GD2    MIME-type?
  GIF
  JPEG
* PNG
  TGA
  TIFF
* WBMP
* WebP
* XBM
  XPM

*/

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    gdImagePtr(*create_from_data)(int,void*);
    gdImagePtr(*create_from_stream)(FILE*);
    cairo_surface_t *surface;
};


static const char **
_gd_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_BMP
        "image/bmp",
        "image/x-bmp",
        "image/x-MS-bmp",
#endif
#ifdef ENABLE_PNG
        "image/png",
#endif
#ifdef ENABLE_WBMP
        "image/vnd.wap.wbmp",
#endif
#ifdef ENABLE_WEBP
        "image/webp",
#endif
#ifdef ENABLE_XBM
        "image/x-xbitmap",
        "image/x-xbmi",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gd_new(const char *mime_type,abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#ifdef ENABLE_BMP
    if(!strcasecmp(mime_type,"image/bmp")|
       !strcasecmp(mime_type,"image/x-bmp")|
       !strcasecmp(mime_type,"image/x-MS-bmp"))
    {
        h->create_from_data = gdImageCreateFromBmpPtr;
        h->create_from_stream = gdImageCreateFromBmp;
    }else
#endif

#ifdef ENABLE_PNG
    if(!strcasecmp(mime_type,"image/png")){
        h->create_from_data = gdImageCreateFromPngPtr;
        h->create_from_stream = gdImageCreateFromPng;
    }else
#endif
#ifdef ENABLE_WBMP
    if(!strcasecmp(mime_type,"image/vnd.wap.wbmp")){
        h->create_from_data = gdImageCreateFromWBMPPtr;
        h->create_from_stream = gdImageCreateFromWBMP;
    }else
#endif
#ifdef ENABLE_WEBP
    if(!strcasecmp(mime_type,"image/webp")){
        h->create_from_data = gdImageCreateFromWebpPtr;
        h->create_from_stream = gdImageCreateFromWebp;
    }else
#endif
#ifdef ENABLE_XBM
    if(!strcasecmp(mime_type,"image/x-xbitmap")|
       !strcasecmp(mime_type,"image/x-xbmi"))
    {
        h->create_from_data = NULL;
        h->create_from_stream = gdImageCreateFromXbm;
    }else
#endif
    {
        free(h);
        return NULL;
    }
    h->info = info;
    return h;
}


static void
_gd_free(abydos_plugin_handle_t *h)
{
    free(h);
}


static cairo_surface_t *
_surface_from_image(gdImagePtr im,int *width, int *height)
{
    cairo_surface_t *surface;
    uint32_t *dst;
    int rowskip;
    int x;
    int y;
    *width = gdImageSX(im);
    *height = gdImageSY(im);
    //DEBUG_p("is true color: %s",gdImageTrueColor(im)?"yes":"no");
    //DEBUG_p("has alpha: %s","?");
    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,*width,*height);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-*width;
    for(y=0;y<*height;++y){
        for(x=0;x<*width;++x){
            int pix = gdImageGetTrueColorPixel(im,x,y);
            int alpha = gdAlphaMax - gdTrueColorGetAlpha(pix);
            *dst++ = ((alpha*gdTrueColorGetRed(pix)/gdAlphaMax)<<16)|
                     ((alpha*gdTrueColorGetGreen(pix)/gdAlphaMax)<<8)|
                     ((alpha*gdTrueColorGetBlue(pix)/gdAlphaMax)<<0)|
                     (255*alpha/gdAlphaMax)<<24;
        }
        dst += rowskip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_gd_create_from_stream(abydos_plugin_handle_t *h,FILE *f){
    gdImagePtr im = h->create_from_stream(f);
    if(!im){
        h->info->error = "Failed to creage GD image";
        return -1;
    }
    h->surface = _surface_from_image(im,&h->info->width,&h->info->height);
    gdImageDestroy(im);
    return 0;
}


static int
_gd_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    if(h->create_from_data){
        gdImagePtr im = h->create_from_data(len,(void*)data);
        if(!im){
            h->info->error = "Failed to creage GD image";
            return -1;
        }
        h->surface = _surface_from_image(im,&h->info->width,&h->info->height);
        return 0;
    }else{
        FILE *f;
        int res;
        f = fmemopen((void*)data,len,"r");
        if(!f)
            return -1;
        res = _gd_create_from_stream(h,f);
        fclose(f);
        return res;
    }
}


static int
_gd_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    FILE *f;
    int res;
    DEBUG_p("filename: %s",filename);
    f = fopen(filename,"r");
    if(!f){
        h->info->error = "Failed open file for reading";
        return -1;
    }
    res = _gd_create_from_stream(h,f);
    fclose(f);
    return res;
}


static cairo_surface_t *
_gd_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *visible_layers,int prefered_variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0(){
    static abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gd_get_mime_types,
        .new = _gd_new,
        .free = _gd_free,
        .create_from_data = _gd_create_from_data,
        .create_from_file = _gd_create_from_file,
        .get_image_surface = _gd_get_image_surface,
    };
    return &p;
}
