/*
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <assert.h>
#include <debug.h>
#include <libheif/heif.h>
#include <stdlib.h>                     /* malloc(), free() */


typedef struct{
    int width;
    int height;
    cairo_surface_t *surface;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    //heif_context *ctx;
    page_t *page;
};


static const char **
_heif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/heic",
        "image/heic-sequence",
        "image/heif",
        "image/heif-sequence",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_heif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    //h->ctx = heif_context_alloc();
    h->page = NULL;
    return h;
}


static void
_heif_free(abydos_plugin_handle_t *h)
{
    //heif_context_free(h->ctx);
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            cairo_surface_destroy(h->page[i].surface);
        free(h->page);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_img(abydos_plugin_handle_t *h, struct heif_image *img,int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;
    const uint8_t* src;
    int src_skip;
    int x;
    int y;
    src = heif_image_get_plane_readonly(img, heif_channel_interleaved, &src_skip);
    src_skip = src_skip/4-width;
    for(y=height;y>0;--y){
        for(x=width;x>0;--x){
            int alpha = src[3];
            *dst++ = (alpha*src[0]/255)<<16|
                     (alpha*src[1]/255)<< 8|
                     (alpha*src[2]/255)<< 0|
                     alpha<<24;
            src += 4;
        }
        src += src_skip;
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_heif_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    struct heif_error err;
    struct heif_context* hc;
    struct heif_image_handle* imgh = NULL;
    heif_item_id *id = NULL;
    struct heif_image* img = NULL;
    int res = -1;
    int i;

    hc = heif_context_alloc();

    err = heif_context_read_from_memory_without_copy(hc, bytes, len, NULL);
    if(err.code != heif_error_Ok) {
        h->info->error = err.message;
        goto cleanup;
    }

    h->info->page_count = heif_context_get_number_of_top_level_images(hc);
    DEBUG_p("images: %d",h->info->page_count);
    h->page = (page_t*)malloc(h->info->page_count * sizeof(page_t));
    id = (heif_item_id*)malloc(h->info->page_count * sizeof(heif_item_id));
    heif_context_get_list_of_top_level_image_IDs(hc, id, h->info->page_count);

    for(i=0;i<h->info->page_count;++i){
        int width;
        int height;
        err = heif_context_get_image_handle(hc, id[i], &imgh);
        if(err.code != heif_error_Ok) {
            h->info->error = err.message;
            goto cleanup;
        }

        err = heif_decode_image(imgh, &img, heif_colorspace_RGB, heif_chroma_interleaved_RGBA, NULL);
        if(err.code != heif_error_Ok) {
            h->info->error = err.message;
            goto cleanup;
        }
        heif_image_handle_release(imgh);
        imgh = NULL;

        h->page[i].width = width = heif_image_get_width(img, heif_channel_interleaved);
        h->page[i].height = height = heif_image_get_height(img, heif_channel_interleaved);
        h->page[i].surface = _surface_from_img(h,img,width,height);
        heif_image_release(img);
        img = NULL;
        if(width>h->info->width)
            h->info->width = width;
        if(height>h->info->height)
            h->info->height = height;
    }


    res = 0;

cleanup:
    if(img)
        heif_image_release(img);
    if(id)
        free(id);
    if(imgh)
        heif_image_handle_release(imgh);
    if(res<0)
        heif_context_free(hc);

    return res;
}

#if 0

    struct heif_error err;
    int num;
    //ContextReleaser cr(ctx);
    err = heif_context_read_from_memory_without_copy(ctx, bytes, len, NULL);
    if(err.code != 0){
        DEBUG_P("could not read HEIF file: %s",err.message);
        return -1;
    }

    num = heif_context_get_number_of_top_level_images(ctx);
    if(num <= 0){
        DEBUG_p("no images");
        return -1;
    }

    DEBUG_p("file contains %d images",num);

    //heif_item_id* image_IDs = (heif_item_id*)alloca(num_images * sizeof(heif_item_id));
    //num_images = heif_context_get_list_of_top_level_image_IDs(ctx, image_IDs, num_images);


    //std::string filename;
    //size_t image_index = 1;  // Image filenames are "1" based.

    for(i=0;i<num;++i){
        struct heif_image_handle *handle;
        err = heif_context_get_image_handle(h->ctx,image_IDs[idx],&handle);
        if (err.code){
            DEBUG_p("could not read HEIF image %d: %s",i,err.message);
            //return -1;
            continue;
        }

        int has_alpha = heif_image_handle_has_alpha_channel(handle);
        struct heif_decoding_options* decode_options = heif_decoding_options_alloc();

        int bit_depth = heif_image_handle_get_luma_bits_per_pixel(handle);

        struct heif_image* image;
        err = heif_decode_image(handle,
                                &image,
                                encoder->colorspace(has_alpha),
                                encoder->chroma(has_alpha, bit_depth),
                                decode_options);
        heif_decoding_options_free(decode_options);
        if(err.code){
            heif_image_handle_release(handle);
            // std::cerr << "Could not decode HEIF image: " << idx << ": " << err.message << "\n";

            h->info->error = "Could not decode HEIF image";
            return -1;
        }

    if (image) {
      bool written = encoder->Encode(handle, image, filename.c_str());
      if (!written) {
        fprintf(stderr,"could not write image\n");
      } else {
        printf("Written to %s\n", filename.c_str());
      }
      heif_image_release(image);


      int has_depth = heif_image_handle_has_depth_image(handle);
      if (has_depth) {
        heif_item_id depth_id;
        int nDepthImages = heif_image_handle_get_list_of_depth_image_IDs(handle, &depth_id, 1);
        assert(nDepthImages==1);
        (void)nDepthImages;

        struct heif_image_handle* depth_handle;
        err = heif_image_handle_get_depth_image_handle(handle, depth_id, &depth_handle);
        if (err.code) {
          heif_image_handle_release(handle);
          std::cerr << "Could not read depth channel\n";
          return 1;
        }

        int bit_depth = 10; // TODO

        struct heif_image* depth_image;
        err = heif_decode_image(depth_handle,
                                &depth_image,
                                encoder->colorspace(false),
                                encoder->chroma(false, bit_depth),
                                nullptr);
        if (err.code) {
          heif_image_handle_release(depth_handle);
          heif_image_handle_release(handle);
          std::cerr << "Could not decode depth image: " << err.message << "\n";
          return 1;
        }

        std::ostringstream s;
        s << output_filename.substr(0, output_filename.find('.'));
        s << "-depth";
        s << output_filename.substr(output_filename.find('.'));

        written = encoder->Encode(depth_handle, depth_image, s.str());
        if (!written) {
          fprintf(stderr,"could not write depth image\n");
        } else {
          printf("Depth image written to %s\n", s.str().c_str());
        }
      }
      heif_image_handle_release(handle);
    }

    image_index++;
  }

    return -1;
}
#endif


static cairo_surface_t *
_heif_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


static void
_heif_get_duration(abydos_plugin_handle_t *h,double *duration)
{
    DEBUG_p("begin");
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _heif_get_mime_types,
        .new = _heif_new,
        .free = _heif_free,
        .create_from_data = _heif_create_from_data,
        .get_image_surface = _heif_get_image_surface,
        .get_duration = _heif_get_duration
    };
    return &p;
}
