/*
 * Copyright (C) 2018 - 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <autoconf.h>

#include <abydos-plugin.h>
#if IMAGEMAGICK_MAJOR_VERSION >= 7
# include <MagickCore/MagickCore.h>
#else
# include <magick/MagickCore.h>
#endif
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

#include "debug.h"

#define ENABLE_AVSX
/* #define ENABLE_CINEON */ /* not working (bad output), not working with convert either */
#define ENABLE_DICOM
/* #define ENABLE_DJVU */ /* not working? */
/* #define ENABLE_EXR */ /* not working (just black) */
/* #define ENABLE_FITS */ /* not working (just black) */
/* #define ENABLE_GIF */
/* #define ENABLE_HEIF */ /* not working (wrong colors), but works with convert */
#define ENABLE_JPEG
#define ENABLE_JPEG2000
/* #define ENABLE_MTV */ /* not working ("unable to read image data") */
#define ENABLE_MNG
#define ENABLE_PCX
#define ENABLE_PHOTOSHOP
/* #define ENABLE_SUNRASTER */ /* not working ("insufficient image data in file") */
#define ENABLE_TARGA
/* #define ENABLE_TIFF */
/* #define ENABLE_XBITMAP */ /* not working (bad output), but works with convert */
#define ENABLE_XPIXMAP

#define QUANTUM_MAX (0xffffffff>>(32-MAGICKCORE_QUANTUM_DEPTH))

/*

Formats supported by imagemagick:

  AAI     | AAI Dune image
  ART     | PFS: 1st Publisher	Format originally used on the Macintosh (MacPaint?) and later used for PFS: 1st Publisher clip art.
  ARW     | Sony Digital Camera Alpha Raw Image Format
  AVI     | Microsoft Audio/Visual Interleaved
* AVS     | AVS X image
  BPG     | Better Portable Graphics	Use -quality to specify the image compression quality. To meet the requirements of BPG, the quality argument divided by 2 (e.g. -quality 92 assigns 46 as the BPG compression.
  BMP     | Microsoft Windows bitmap	By default the BMP format is version 4. Use BMP3 and BMP2 to write versions 3 and 2 respectively.
  CALS    | Continuous Acquisition and Life-cycle Support Type 1 image	Specified in MIL-R-28002 and MIL-PRF-28002. Standard blueprint archive format as used by the US military to replace microfiche.
  CGM     | Computer Graphics Metafile	Requires ralcgm to render CGM files.
* CIN     | Kodak Cineon Image Format	Use -set to specify the image gamma or black and white points (e.g. -set gamma 1.7, -set reference-black 95, -set reference-white 685). Properties include cin:file.create_date, cin:file.create_time, cin:file.filename, cin:file.version, cin:film.count, cin:film.format, cin:film.frame_id, cin:film.frame_position, cin:film.frame_rate, cin:film.id, cin:film.offset, cin:film.prefix, cin:film.slate_info, cin:film.type, cin:image.label, cin:origination.create_date, cin:origination.create_time, cin:origination.device, cin:origination.filename, cin:origination.model, cin:origination.serial, cin:origination.x_offset, cin:origination.x_pitch, cin:origination.y_offset, cin:origination.y_pitch, cin:user.data.
  CR2     | Canon Digital Camera Raw Image Format	Requires an explicit image format otherwise the image is interpreted as a TIFF image (e.g. cr2:image.cr2).
  CRW     | Canon Digital Camera Raw Image Format
  CUBE    | Cube Color lookup table converted to a HALD image	Select levels like this: cube:Vibrant.cube[8] for level 8
  CUR     | Microsoft Cursor Icon
  CUT     | DR Halo
* DCM     | Digital Imaging and Communications in Medicine (DICOM) image
  DCR     | Kodak Digital Camera Raw Image File
  DCX     | ZSoft IBM PC multi-page Paintbrush image
  DDS     | Microsoft Direct Draw Surface	Use -define to specify the compression (e.g. -define dds:compression={dxt1, dxt5, none}). Other defines include dds:cluster-fit={true,false}, dds:weight-by-alpha={true,false}, dds:fast-mipmaps={true,false}, and use dds:mipmaps to set the number of mipmaps (use fromlist to use the image list).
  DIB     | Microsoft Windows Device Independent Bitmap	DIB is a BMP file without the BMP header. Used to support embedded images in compound formats like WMF.
* DJVU    |
  DNG     | Digital Negative	Requires an explicit image format otherwise the image is interpreted as a TIFF image (e.g. dng:image.dng).
  DOT     | Graph Visualization	Use -define to specify the layout engine (e.g. -define dot:layout-engine=twopi).
  DPX     | SMPTE Digital Moving Picture Exchange 2.0 (SMPTE 268M-2003)	Use -set to specify the image gamma or black and white points (e.g. -set gamma 1.7, -set reference-black 95, -set reference-white 685).
  EPDF    | Encapsulated Portable Document Format
  EPI     | Adobe Encapsulated PostScript Interchange format	Requires Ghostscript to read.
  EPS     | Adobe Encapsulated PostScript	Requires Ghostscript to read.
  EPSF    | Adobe Encapsulated PostScript	Requires Ghostscript to read.
  EPSI    | Adobe Encapsulated PostScript Interchange format	Requires Ghostscript to read.
  EPT     | Adobe Encapsulated PostScript Interchange format with TIFF preview	Requires Ghostscript to read.
* EXR     | High dynamic-range (HDR) file format developed by Industrial Light & Magic	See High Dynamic-Range Images for details on this image format. To specify the output color type, use -define exr:color-type={RGB,RGBA,YC,YCA,Y,YA,R,G,B,A}. Use -sampling-factor to specify the sampling rate for YC(A) (e.g. 2x2 or 4:2:0). Requires the OpenEXR delegate library.
  FAX     | Group 3 TIFF	This format is a fixed width of 1728 as required by the standard. See TIFF format. Note that FAX machines use non-square pixels which are 1.5 times wider than they are tall but computer displays use square pixels so FAX images may appear to be narrow unless they are explicitly resized using a resize specification of 100x150%.
  FIG     | FIG graphics format	Requires TransFig.
* FITS    | Flexible Image Transport System
  FPX     | FlashPix Format	FlashPix has the option to store mega- and giga-pixel images at various resolutions in a single file which permits conservative bandwidth and fast reveal times when displayed within a Web browser. Requires the FlashPix SDK. Specify the FlashPix viewing parameters with the -define fpx:view.
* GIF     | CompuServe Graphics Interchange Format	8-bit RGB PseudoColor with up to 256 palette entries. Specify the format GIF87 to write the older version 87a of the format. Use -transparent-color to specify the GIF transparent color (e.g. -transparent-color wheat).
  GPLT    | Gnuplot plot files	Requires gnuplot4.0.tar.Z or later.
  HDR     | Radiance RGBE image format
  HDR     | Radiance RGBE image format
* HEIC    | Apple High efficiency Image Format
  HPGL    | HP-GL plotter language	Requires hp2xx-3.4.4.tar.gz
  HRZ     | Slow Scane TeleVision
  ICO     | Microsoft Windows icon
  JBIG    | Joint Bi-level Image experts Group file interchange format	Also known as BIE and JBG. Requires jbigkit-1.6.tar.gz.
  JNG     | Multiple-image Network Graphics	JPEG in a PNG-style wrapper with transparency. Requires libjpeg and libpng-1.0.11 or later, libpng-1.2.5 or later recommended.
* JP2     | JPEG-2000 JP2 File Format Syntax	Specify the encoding options with the -define option. See JP2 Encoding Options for more details.
  JPT     | JPEG-2000 Code Stream Syntax	Specify the encoding options with the -define option See JP2 Encoding Options for more details.
  J2C     | JPEG-2000 Code Stream Syntax	Specify the encoding options with the -define option See JP2 Encoding Options for more details.
  J2K     | JPEG-2000 Code Stream Syntax	Specify the encoding options with the -define option See JP2 Encoding Options for more details.
* JPEG    | Joint Photographic Experts Group JFIF format
  JXR     | JPEG extended range	Requires the jxrlib delegate library. Put the JxrDecApp and JxrEncApp applications in your execution path.
  MAN     | Unix reference manual pages	Requires that GNU groff and Ghostcript are installed.
  MAT     | MATLAB image format
  MIFF    | Magick image file format	This format persists all image attributes known to ImageMagick. To specify a single precision floating-point format, use -define quantum:format=floating-point. Set the depth to 32 for single precision floats, 64 for double precision, and 16 for half-precision.
  MONO    | Bi-level bitmap in least-significant-byte first order
* MNG     | Multiple-image Network Graphics	A PNG-like Image Format Supporting Multiple Images, Animation and Transparent JPEG. Requires libpng-1.0.11 or later, libpng-1.2.5 or later recommended. An interframe delay of 0 generates one frame with each additional layer composited on top. For motion, be sure to specify a non-zero delay.
  M2V     | Motion Picture Experts Group file interchange format (version 2)	Requires ffmpeg.
  MPEG    | Motion Picture Experts Group file interchange format (version 1)	Requires ffmpeg.
  MPC     | Magick Persistent Cache image file format	The most efficient data processing pattern is a write-once, read-many-times pattern. The image is generated or copied from source, then various analyses are performed on the image pixels over time. MPC supports this pattern. MPC is the native in-memory ImageMagick uncompressed file format. This file format is identical to that used by ImageMagick to represent images in memory and is read by mapping the file directly into memory. The MPC format is not portable and is not suitable as an archive format. It is suitable as an intermediate format for high-performance image processing. The MPC format requires two files to support one image. Image attributes are written to a file with the extension .mpc, whereas, image pixels are written to a file with the extension .cache.
  MRW     | Sony (Minolta) Raw Image File
  MSL     | Magick Scripting Language	MSL is the XML-based scripting language supported by the conjure utility. MSL requires the libxml2 delegate library.
* MTV     | MTV Raytracing image format
  MVG     | Magick Vector Graphics.	The native ImageMagick vector metafile format. A text file containing vector drawing commands accepted by convert's -draw option.
  NEF     | Nikon Digital SLR Camera Raw Image File
  ORF     | Olympus Digital Camera Raw Image File
  OTB     | On-the-air Bitmap
  P7      | Xv's Visual Schnauzer thumbnail format
  PALM    | Palm pixmap
  PBM     | Portable bitmap format (black and white)
  PCD     | Photo CD	The maximum resolution written is 768x512 pixels since larger images require huffman compression (which is not supported). Use -bordercolor to specify the border color (e.g. -bordercolor black).
  PCDS    | Photo CD	Decode with the sRGB color tables.
* PCX     | ZSoft IBM PC Paintbrush file
  PDB     | Palm Database ImageViewer Format
  PDF     | Portable Document Format
  PEF     | Pentax Electronic File	Requires an explicit image format otherwise the image is interpreted as a TIFF image (e.g. pef:image.pef).
  PES     | Embrid Embroidery Format
  PFA     | Postscript Type 1 font (ASCII)	Opening as file returns a preview image.
  PFB     | Postscript Type 1 font (binary)	Opening as file returns a preview image.
  PFM     | Portable float map format
  PGM     | Portable graymap format (gray scale)
  PICON   | Personal Icon
  PICT    | Apple Macintosh QuickDraw/PICT file
  PIX     | Alias/Wavefront RLE image format
  PNG     | Portable Network Graphics	Requires libpng-1.0.11 or later, libpng-1.2.5 or later recommended. The PNG specification does not support pixels-per-inch units, only pixels-per-centimeter. To avoid reading a particular associated image profile, use -define profile:skip=name (e.g. profile:skip=ICC).
  PNM     | Portable anymap	PNM is a family of formats supporting portable bitmaps (PBM) , graymaps (PGM), and pixmaps (PPM). There is no file format associated with pnm itself. If PNM is used as the output format specifier, then ImageMagick automagically selects the most appropriate format to represent the image. The default is to write the binary version of the formats. Use -compress none to write the ASCII version of the formats.
  PPM     | Portable pixmap format (color)
  PS      | Adobe PostScript file	Requires Ghostscript to read. To force ImageMagick to respect the crop box, use -define (e.g. -define eps:use-cropbox=true). Use -density to improve the appearance of your Postscript rendering (e.g. -density 300x300). Use -alpha remove to remove transparency. To specify direct conversion from PDF to Postscript, use -define delegate:bimodel=true.
  PS2     | Adobe Level II PostScript file	Requires Ghostscript to read.
  PS3     | Adobe Level III PostScript file	Requires Ghostscript to read.
  PSB     | Adobe Large Document Format
* PSD     | Adobe Photoshop bitmap file	Use -define psd:alpha-unblend=off to disable alpha blenning in the merged image. Use -define psd:additional-info=all|selective to transfer additional information from the input PSD file to output PSD file. The 'selective' option will preserve all additional information that is not related to the geometry of the image. The 'all' option should only be used when the geometry of the image has not been changed. This option is helpful when transferring non-simple layers, such as adjustment layers from the input PSD file to the output PSD file. This define is available as of Imagemagick version 6.9.5-8. Use -define psd:preserve-opacity-mask=true to preserve the opacity mask of a layer and add it back to the layer when the image is saved.
  PTIF    | Pyramid encoded TIFF	Multi-resolution TIFF containing successively smaller versions of the image down to the size of an icon.
  PWP     | Seattle File Works multi-image file
  RAD     | Radiance image file	Requires that ra_ppm from the Radiance software package be installed.
  RAF     | Fuji CCD-RAW Graphic File
  RGF     | LEGO Mindstorms EV3 Robot Graphics File
  RLA     | Alias/Wavefront image file
  RLE     | Utah Run length encoded image file
  SCT     | Scitex Continuous Tone Picture
  SFW     | Seattle File Works image
  SGI     | Irix RGB image
  SID, MrSID	Multiresolution seamless image	Requires the mrsidgeodecode command line utility that decompresses MG2 or MG3 SID image files.
* SUN     | SUN Rasterfile
  SVG     | Scalable Vector Graphics	ImageMagick utilizes inkscape if its in your execution path otherwise RSVG. If neither are available, ImageMagick reverts to its internal SVG renderer. The default resolution is 96 DPI. Use -size command line option to specify the maximum width and height.
  TEXT    | text file	Requires an explicit format specifier to read, e.g. text:README.txt.
* TGA     | Truevision Targa image	Also known as formats ICB, VDA, and VST.
* TIFF    | Tagged Image File Format	Also known as TIF. Requires tiff-v3.6.1.tar.gz or later. Use -define to specify the rows per strip (e.g. -define tiff:rows-per-strip=8). To define the tile geometry, use for example, -define tiff:tile-geometry=128x128. To specify a signed format, use -define quantum:format=signed. To specify a single-precision floating-point format, use -define quantum:format=floating-point. Set the depth to 64 for a double-precision floating-point format. Use -define quantum:polarity=min-is-black or -define quantum:polarity=min-is-white toggle the photometric interpretation for a bilevel image. Specify the extra samples as associated or unassociated alpha with, for example, -define tiff:alpha=unassociated. Set the fill order with -define tiff:fill-order=msb|lsb. Set the TIFF endianess with -define tiff:endian=msb|lsb. Use -define tiff:exif-properties=false to skip reading the EXIF properties. You can set a number of TIFF software attributes including document name, host computer, artist, timestamp, make, model, software, and copyright. For example, -set tiff:software "My Company". If you want to ignore certain TIFF tags, use this option: -define tiff:ignore-tags=comma-separated-list-of-tag-IDs. Since version 6.9.1-4 there is support for reading photoshop layers in TIFF files, this can be disabled with -define tiff:ignore-layers=true
  TIM     | PSX TIM file
  TTF     | TrueType font file	Requires freetype 2. Opening as file returns a preview image. Use -set if you do not want to hint glyph outlines after their scaling to device pixels (e.g. -set type:hinting off).
  UYVY    | Interleaved YUV raw image	Use -size and -depth command line options to specify width and height. Use -sampling-factor to set the desired subsampling (e.g. -sampling-factor 4:2:2).
  VICAR   | VICAR rasterfile format
  VIFF    | Khoros Visualization Image File Format
  WBMP    | Wireless bitmap 	Support for uncompressed monochrome only.
  WDP     | JPEG extended range	Requires the jxrlib delegate library. Put the JxrDecApp and JxrEncApp applications in your execution path.
  WEBP    | Weppy image format	Requires the WEBP delegate library. Specify the encoding options with the -define option See WebP Encoding Options for more details.
  WMF     | Windows Metafile	Requires libwmf. By default, renders WMF files using the dimensions specified by the metafile header. Use the -density option to adjust the output resolution, and thereby adjust the output size. The default output resolution is 72DPI so -density 144 results in an image twice as large as the default. Use -background color to specify the WMF background color (default white) or -texture filename to specify a background texture image.
  WPG     | Word Perfect Graphics File
* XBM     | X Windows system bitmap, black and white only	Used by the X Windows System to store monochrome icons.
  XCF     | GIMP image
* XPM     | X Windows system pixmap	Also known as PM. Used by the X Windows System to store color icons.
  XWD     | X Windows system window dump
  X3F     | Sigma Camera RAW Picture File

*/

typedef struct{
    cairo_surface_t *surface;
    double duration;
}multi_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    const char *magick_type;
    enum{
        MT_NONE,
        MT_PAGE,
        MT_LAYER,
        MT_FRAME
    }multi_type;
    multi_t *multi;
    int multi_count;
};


static void
_magick_done()
{
    MagickCoreTerminus();
}


static const char **
_magick_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_AVSX
        "image/x-avsx",
#endif
#ifdef ENABLE_CINEON
        "image/x-cineon",
#endif
#ifdef ENABLE_DICOM
        "application/dicom",
#endif
#ifdef ENABLE_DJVU
        "image/vnd.djvu",
#endif
#ifdef ENABLE_EXR
        "image/x-exr",
#endif
#ifdef ENABLE_FITS
        "application/fits",
        "image/fits",
#endif
#ifdef ENABLE_GIF
        "image/gif",
#endif
#ifdef ENABLE_HEIF
        "image/heic",
        "image/heic-sequence",
        "image/heif",
        "image/heif-sequence",
#endif
#ifdef ENABLE_JPEG
        "image/jpeg",
#endif
#ifdef ENABLE_JPEG2000
        "image/jp2",
        "image/jpeg2000",
        "image/jpx",
#endif
#ifdef ENABLE_MNG
        "video/x-mng",
#endif
#ifdef ENABLE_MTV
        "image/x-mtv",
#endif
#ifdef ENABLE_PCX
        "image/vnd.zbrush.pcx",
        "image/x-pcx",
#endif
#ifdef ENABLE_PHOTOSHOP
        "image/vnd.adobe.photoshop",
#endif
#ifdef ENABLE_SUNRASTER
        "image/x-sun-raster",
#endif
#ifdef ENABLE_TARGA
        "image/x-tga",
#endif
#ifdef ENABLE_TIFF
        "image/tiff",
#endif
#ifdef ENABLE_XBITMAP
        "image/x-xbitmap",
        "image/x-xbmi",
#endif
#ifdef ENABLE_XPIXMAP
        "image/x-xpixmap",
        "image/x-xpmi",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_magick_new(const char *mime_type,abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->multi_type = MT_NONE;
#ifdef ENABLE_AVSX
    if(!strcasecmp(mime_type,"image/x-avsx")){
        h->magick_type = "AVS";
    }else
#endif
#ifdef ENABLE_CINEON
    if(!strcasecmp(mime_type,"image/x-cineon")){
        h->magick_type = "CIN";
    }else
#endif
#ifdef ENABLE_DICOM
    if(!strcasecmp(mime_type,"application/dicom")){
        h->magick_type = "DCM";
    }else
#endif
#ifdef ENABLE_DJVU
    if(!strcasecmp(mime_type,"image/vnd.djvu")){
        h->magick_type = "DJVU";
        h->multi_type = MT_PAGE;
    }else
#endif
#ifdef ENABLE_EXR
    if(!strcasecmp(mime_type,"image/x-exr")){
        h->magick_type = "EXR";
    }else
#endif
#ifdef ENABLE_FITS
    if(!strcasecmp(mime_type,"application/fits")||
       !strcasecmp(mime_type,"image/fits"))
    {
        h->magick_type = "FITS";
    }else
#endif
#ifdef ENABLE_GIF
    if(!strcasecmp(mime_type,"image/gif")){
        h->magick_type = "GIF";
        h->multi_type = MT_FRAME;
    }else
#endif
#ifdef ENABLE_HEIF
    if(!strcasecmp(mime_type,"image/heic")||
       !strcasecmp(mime_type,"image/heic-sequence")||
       !strcasecmp(mime_type,"image/heif")||
       !strcasecmp(mime_type,"image/heif-sequence"))
    {
        h->magick_type = "HEIC";
    }else
#endif
#ifdef ENABLE_JPEG
    if(!strcasecmp(mime_type,"image/jpeg")){
        h->magick_type = "JPEG";
    }else
#endif
#ifdef ENABLE_JPEG2000
    if(!strcasecmp(mime_type,"image/jp2")||
       !strcasecmp(mime_type,"image/jpeg2000")||
       !strcasecmp(mime_type,"image/jpx"))
    {
        h->magick_type = "JP2";
    }else
#endif
#ifdef ENABLE_MNG
    if(!strcasecmp(mime_type,"video/x-mng")){
        h->magick_type = "MNG";
        h->multi_type = MT_FRAME;
    }else
#endif
#ifdef ENABLE_MTV
    if(!strcasecmp(mime_type,"image/x-mtv")){
        h->magick_type = "MTV";
    }else
#endif
#ifdef ENABLE_PCX
    if(!strcasecmp(mime_type,"image/vnd.zbrush.pcx")||
       !strcasecmp(mime_type,"image/x-pcx"))
    {
        h->magick_type = "PCX";
    }else
#endif
#ifdef ENABLE_PHOTOSHOP
    if(!strcasecmp(mime_type,"image/vnd.adobe.photoshop")){
        h->magick_type = "PSD";
        h->multi_type = MT_LAYER;
    }else
#endif
#ifdef ENABLE_SUNRASTER
    if(!strcasecmp(mime_type,"image/x-sun-raster")){
        h->magick_type = "SUN";
    }else
#endif
#ifdef ENABLE_TARGA
    if(!strcasecmp(mime_type,"image/x-tga")){
        h->magick_type = "TGA";
    }else
#endif
#ifdef ENABLE_TIFF
    if(!strcasecmp(mime_type,"image/tiff")){
        h->magick_type = "TIFF";
        h->multi_type = MT_PAGE;
    }else
#endif
#ifdef ENABLE_XBITMAP
    if(!strcasecmp(mime_type,"image/x-xbitmap")||
       !strcasecmp(mime_type,"image/x-xbmi"))
    {
        h->magick_type = "XBM";
    }else
#endif
#ifdef ENABLE_XPIXMAP
    if(!strcasecmp(mime_type,"image/x-xpixmap")||
       !strcasecmp(mime_type,"image/x-xpmi"))
    {
        h->magick_type = "XPM";
    }else
#endif
    {
        free(h);
        return NULL;
    }
    h->info = info;
    h->multi = NULL;
    h->multi_count = 0;
    return h;
}


static void
_magick_free(abydos_plugin_handle_t *h)
{
    if(h->multi){
        int i;
        for(i=0;i<h->multi_count;++i)
            cairo_surface_destroy(h->multi[i].surface);
        free(h->multi);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_image(Image *image,ExceptionInfo *exception)
{
    cairo_surface_t *surface;
#if IMAGEMAGICK_MAJOR_VERSION >= 7
    const Quantum *src;
    int channels;
#else
    PixelPacket *src;
#endif
    uint32_t *dst;
    int rowskip;
    int x;
    int y;


#if MagickLibVersion >= 0x640
    src = GetVirtualPixels(image,0,0,image->columns,image->rows,exception);
#else
    src = (PixelPacket*)AcquireImagePixels(image,0,0,image->columns,image->rows,exception);
#endif
    if(exception->severity){
        /* Can't get the pixels for the sub image, strange but not fatal. */
        DEBUG_p("failed to get sub image");
        ClearMagickException(exception);
        return NULL;
    }

#if IMAGEMAGICK_MAJOR_VERSION >= 7
    channels = GetPixelChannels(image);
    DEBUG_p("channels: %d",channels);
#endif

    if(GetImageAlphaChannel(image)){
        surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,image->columns,image->rows);
    }else{
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,image->columns,image->rows);
    }

    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - image->columns;;

#if IMAGEMAGICK_MAJOR_VERSION >= 7
    for(y=0;y<image->rows;++y){
        for(x=0;x<image->columns;++x){
            int alpha = GetPixelAlpha(image,src);
            int r = alpha*GetPixelRed(image,src)/QUANTUM_MAX;
            int g = alpha*GetPixelGreen(image,src)/QUANTUM_MAX;
            int b = alpha*GetPixelBlue(image,src)/QUANTUM_MAX;
            *dst++ = (r>>(MAGICKCORE_QUANTUM_DEPTH-8))<<16|
                     (g>>(MAGICKCORE_QUANTUM_DEPTH-8))<< 8|
                     (b>>(MAGICKCORE_QUANTUM_DEPTH-8))<< 0|
                     (alpha>>(MAGICKCORE_QUANTUM_DEPTH-8))<<24;
            src += channels;
        }
        dst += rowskip;
    }
#else
    if(GetImageAlphaChannel(image)){
        for(y=0;y<image->rows;++y){
            for(x=0;x<image->columns;++x){
                *dst++ = ((src->red>>(MAGICKCORE_QUANTUM_DEPTH-8))<<16)
                        |((src->green>>(MAGICKCORE_QUANTUM_DEPTH-8))<<8)
                        |((src->blue>>(MAGICKCORE_QUANTUM_DEPTH-8))<<0)
                        |((255-(src->opacity>>(MAGICKCORE_QUANTUM_DEPTH-8)))<<24);
                ++src;
            }
            dst += rowskip;
        }
    }else{
        for(y=0;y<image->rows;++y){
            for(x=0;x<image->columns;++x){
                *dst++ = ((src->red>>(MAGICKCORE_QUANTUM_DEPTH-8))<<16)
                        |((src->green>>(MAGICKCORE_QUANTUM_DEPTH-8))<<8)
                        |((src->blue>>(MAGICKCORE_QUANTUM_DEPTH-8))<<0);
                ++src;
            }
            dst += rowskip;
        }
    }
#endif
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_magick_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    ExceptionInfo *exception = AcquireExceptionInfo();
    ImageInfo *info = CloneImageInfo(NULL);
    Image *image = NULL;
    int ret = -1;
    int i;

    strcpy(info->magick,h->magick_type);
    image = BlobToImage(info,data,len,exception);
    if(exception->severity){
        DEBUG_p("invalid image");
        h->info->error = "Invalid image";
        goto out;
    }
    h->info->width = image->columns;
    h->info->height = image->rows;
    h->multi_count = GetImageListLength(image);
    DestroyImageInfo(info);

    DEBUG_p("multi count: %d",h->multi_count);
    switch(h->multi_type){
    case MT_NONE:
        if(h->multi_count>1)
            DEBUG_p("multiple sub images was not expected for an image of type %s",h->magick_type);
        break;
    case MT_PAGE:
        h->info->page_count = h->multi_count;
        break;
    case MT_LAYER:
        h->info->layer_count = h->multi_count;
        break;
    case MT_FRAME:
        h->info->frame_count = h->multi_count;
        break;
    }

    h->multi = (multi_t*)malloc(h->multi_count * sizeof(multi_t));
    for(i=0;i<h->multi_count;++i){
        Image *subimage = GetImageFromList(image,i);
        h->multi[i].surface = _surface_from_image(subimage,exception);
        if(MT_FRAME == h->multi_type && subimage->ticks_per_second > 0){
            h->multi[i].duration = (double)subimage->delay / subimage->ticks_per_second;
        }else{
            h->multi[i].duration = -1;
        }
    }
    ret = 0;
out:
    if(image)
        DestroyImage(image);
    if(exception)
        DestroyExceptionInfo(exception);
    return ret;
}


static cairo_surface_t *
_magick_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *visible_layers,int prefered_variant,double time,int frame)
{
    int index;
    switch(h->multi_type){
    case MT_PAGE:
        index = page;
        break;
    case MT_FRAME:
        index = frame;
        break;
    default:
        index = 0;
        break;
    }
    return cairo_surface_reference(h->multi[index].surface);
}


static void
_magick_get_duration(abydos_plugin_handle_t *h, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->multi[i].duration;
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .done = _magick_done,
        .get_mime_types = _magick_get_mime_types,
        .new = _magick_new,
        .free = _magick_free,
        .create_from_data = _magick_create_from_data,
        .get_image_surface = _magick_get_image_surface,
        .get_duration = _magick_get_duration
    };
    //if(!IsMagickInstantiated())
        MagickCoreGenesis(NULL,0);
    return &p;
}
