/*
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strncasecmp() */

/*
  Can contain:

    Animation           'rle '
    Apple Pixlet Video  'pxlt'
    Apple VC H.263      'h263'
    DV - PAL            'dvcp'
    DV/DVCPRO - NTSC    'dvc '
    DVCPRO - PAL        'dvpp'
    DVCPRO50 - NTSC     'dv5n'
    DVCPRO50 - PAL      'dv5p'
    Foto - JPEG         'jpeg'
    H.264               'avc1'
    None                'raw '
    JPEG 2000           'mjp2'
    MPEG-4 Video        'mp4v'
    PNG                 'png '

*/

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    abydos_t *ar;
};


static const char **
_qtif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/qtif",
        "image/x-quicktime",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_qtif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->ar = NULL;
    return h;
}


static void
_qtif_free(abydos_plugin_handle_t *h)
{
    if(h->ar)
        abydos_destroy(h->ar);
    free(h);
}


#define _get16(p) be16toh(*(uint16_t*)(p))
#define _get32(p) be32toh(*(uint32_t*)(p))


static const char *
_find_atom(uint32_t id,const char *p,size_t len){
    while(len>8){
        size_t atomsize = _get32(p);
        if(atomsize>len)
            return NULL;
        if(*(uint32_t*)(p+4)==id)
            return p;
        p += atomsize;
        len -= atomsize;
    }
    return NULL;
}


static int
_qtif_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    const char* p;
    size_t atomsize;
    size_t datasize;
    const char* mime_type;
    /* A QTIF file cannot possibly be smaller than 25 bytes */
    if(len<25)
        return -1;
    p = _find_atom(*(uint32_t*)"idsc",bytes,len);
    if(!p)
        return -1;
    atomsize = _get32(p+0);
    if(atomsize<16)
        return -1;
    DEBUG_p("content type: '%4s'",p+12);
    if(!strncasecmp(p+12,"jpeg",4)){
        mime_type = "image/jpeg";
    }else if(!strncasecmp(p+12,"mjp2",4)){
        mime_type = "image/jp2";
    }else if(!strncasecmp(p+12,"png ",4)){
        mime_type = "image/png";
    }else{
        /* unknown data type */
        return -1;
    }
    if(atomsize>=44){
        int val1 = _get16(p+40);
        int val2 = _get16(p+42);
        if(val1 > 0 && val2 > 0){
            h->info->width = val1;
            h->info->height = val2;
        }
        if(atomsize>=52){
            val1 = _get32(p+44);
            val2 = _get32(p+48);
            if(val1 > 0 && val2 > 0){
                h->info->pixel_ratio = (double)val1/val2;
            }
        }
    }
    datasize = atomsize>=56 ? _get32(p+52) : 0;

    p = _find_atom(*(uint32_t*)"idat",bytes,len);
    if(!p)
        return -1;
    if(datasize > 0){
        atomsize = _get32(p+0);
        if(datasize != atomsize-8)
            return -1;
    }

    h->ar = abydos_create(mime_type);
    if(!h->ar)
        return -1;
    return abydos_from_data(h->ar,p+8,atomsize-8);
}


static void
_qtif_render(abydos_plugin_handle_t *h,
             cairo_t *cr,
             int page,
             cairo_bool_t *visible_layers,
             int variant,
             double time,
             int frame)
{
    abydos_render(h->ar, cr, 0);
}


static cairo_surface_t *
_qtif_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return abydos_get_image_surface(h->ar, 0);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _qtif_get_mime_types,
        .new = _qtif_new,
        .free = _qtif_free,
        .create_from_data = _qtif_create_from_data,
        .render = _qtif_render,
        .get_image_surface = _qtif_get_image_surface
    };
    return &p;
}
