/*
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>

#include "cons_buf.h"
#include "teletext_dec.h"

/* #define TEST_RENDER */

#define SCREEN_WIDTH 40
#define SCREEN_HEIGHT 25
#define CHAR_WIDTH 12
#define CHAR_HEIGHT 10

static
#include "teletext-12x10.c"

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    Console *cons;
    TeletextDecoder *dec;
#ifndef TEST_RENDER
    cairo_surface_t *surface;
#endif
};


static const char **
_tt_get_mime_types()
{
    static const char *mime_types[] = {
        "text/x-raw-teletext",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_tt_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->cons = cons_buf_new();
    cons_buf_set_size(h->cons,SCREEN_WIDTH,SCREEN_HEIGHT);
    h->dec = teletext_decoder_new(&cons_buf_if,h->cons);
    return h;
}


static void
_tt_free(abydos_plugin_handle_t *h)
{
    teletext_decoder_free(h->dec);
    cons_buf_free(h->cons);
    free(h);
}


static const uint32_t _teletext_palette[8] = {
    0x00000000,    /* black */
    0x00ff0000,    /* red */
    0x0000ff00,    /* green */
    0x00ffff00,    /* yellow */
    0x000000ff,    /* blue */
    0x00ff00ff,    /* magenta */
    0x0000ffff,    /* cyan */
    0x00ffffff     /* white */
};


static void
_draw_rect(uint32_t *dst,int stride,int width,int height,uint32_t col)
{
    while(height-->0){
        int x;
        for(x=0;x<width;++x)
            dst[x] = col;
        dst += stride;
    }
}


static void
_draw_graphic(uint32_t *dst,int stride,int ch,uint32_t bg,uint32_t fg)
{
    _draw_rect(dst           ,stride,6,3,ch&0x01?fg:bg);
    _draw_rect(dst+6         ,stride,6,3,ch&0x02?fg:bg);
    _draw_rect(dst  +stride*3,stride,6,4,ch&0x04?fg:bg);
    _draw_rect(dst+6+stride*3,stride,6,4,ch&0x08?fg:bg);
    _draw_rect(dst  +stride*7,stride,6,3,ch&0x10?fg:bg);
    _draw_rect(dst+6+stride*7,stride,6,3,ch&0x40?fg:bg);
}


static void
_draw_character(uint32_t *dst,int stride,int ch,uint32_t bg,uint32_t fg)
{
    uint16_t *fontchar = &teletext_12x10[(ch-' ')*CHAR_HEIGHT];
    int x;
    int y;
    for(y=0;y<CHAR_HEIGHT;++y){
        int pixels = *fontchar++;
        for(x=0;x<CHAR_WIDTH;++x){
            pixels <<= 1;
            dst[x] = pixels & (1<<CHAR_WIDTH) ? fg : bg;
        }
        dst += stride;
    }
}


static cairo_surface_t *
_surface_from_console(Console *cons)
{
    const ConsoleCell *src = cons_buf_buffer(cons);
    int width = cons_buf_width(cons);
    int height = cons_buf_height(cons);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width*CHAR_WIDTH,height*CHAR_HEIGHT);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int x;
    int y;
    for(y=0;y<SCREEN_HEIGHT;++y){
        for(x=0;x<SCREEN_WIDTH;++x){
            union{
                TeletextState state;
                int attr;
            }s;
            ConsoleCell cell = *src++;
            uint32_t bg;
            uint32_t fg;
            s.attr = cell.attr;
            bg = _teletext_palette[s.state.bg];
            fg = _teletext_palette[s.state.fg];
            if(s.state.graphic && (cell.ch&0x60) != 0x40){
                _draw_graphic(dst+x*CHAR_WIDTH,stride,cell.ch,bg,fg);
            }else{
                _draw_character(dst+x*CHAR_WIDTH,stride,cell.ch,bg,fg);
            }
        }
        dst += stride*CHAR_HEIGHT;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_tt_create_from_data(abydos_plugin_handle_t *h, const char *data,size_t len)
{
    union{
        TeletextState state;
        int attr;
    }s;
    h->info->width = SCREEN_WIDTH * CHAR_WIDTH;
    h->info->height = SCREEN_HEIGHT * CHAR_HEIGHT;
    h->info->pixel_ratio = 25.0/36;

    teletext_reset(&s.state);

#if 1
    while(len--)
        teletext_decoder_write(h->dec,*data++);
#elif 1
    teletext_decoder_write(h->dec,'A');
#else
    cons_buf_charxya(h->cons,0,0,s.attr,'A');
#endif

#ifndef TEST_RENDER
    h->surface = _surface_from_console(h->cons);
#endif

    return 0;
}


static void
_set_color(cairo_t *cr, int idx)
{
    static const struct{
        double r;
        double g;
        double b;
    }color[8] = {
        {0,0,0}, /* black */
        {1,0,0}, /* red */
        {0,1,0}, /* green */
        {1,1,0}, /* yellow */
        {0,0,1}, /* blue */
        {1,0,1}, /* magenta */
        {0,1,1}, /* cyan */
        {1,1,1}  /* white */
    };
    cairo_set_source_rgb(cr,color[idx].r,color[idx].g,color[idx].b);
}




#if 0
static void
_draw_character(cairo_t *cr,int cx,int cy,int ch)
{
    uint16_t *fontchar = &teletext_12x10[(ch-' ')*CHAR_HEIGHT];
    int x;
    int y;
    for(y=0;y<CHAR_HEIGHT;++y){
        int pixels = *fontchar++;
        for(x=0;x<CHAR_WIDTH;++x){
            pixels <<= 1;
            if(pixels & (1<<CHAR_WIDTH))
                cairo_rectangle(cr,cx+x,cy+y,1,1);
        }
    }
}
#endif


#if 0
static void
_draw_graphic(cairo_t *cr,int cx,int cy,int ch)
{
    if(ch&0x01)
        cairo_rectangle(cr,cx+0,cy+0,6,3);
    if(ch&0x02)
        cairo_rectangle(cr,cx+6,cy+0,6,3);
    if(ch&0x04)
        cairo_rectangle(cr,cx+0,cy+3,6,4);
    if(ch&0x08)
        cairo_rectangle(cr,cx+6,cy+3,6,4);
    if(ch&0x10)
        cairo_rectangle(cr,cx+0,cy+7,6,3);
    if(ch&0x40)
        cairo_rectangle(cr,cx+6,cy+7,6,3);
}
#endif


#if 0
static void
_draw_screen(cairo_t *cr,Console *cons)
{
    const ConsoleCell *src = cons_buf_buffer(cons);
    int x;
    int y;
    cairo_save(cr);
    for(y=0;y<SCREEN_HEIGHT;++y){
        for(x=0;x<SCREEN_WIDTH;++x){
            union{
                TeletextState state;
                int attr;
            }s;
            ConsoleCell cell = *src++;
            s.attr = cell.attr;
            _set_color(cr,s.state.bg);
            cairo_rectangle(cr,x*CHAR_WIDTH,y*CHAR_HEIGHT,CHAR_WIDTH,CHAR_HEIGHT);
            cairo_fill(cr);
            if(cell.ch>' '){
                _set_color(cr,s.state.fg);
                if(s.state.graphic && (cell.ch&0x60) != 0x40){
                    _draw_graphic(cr,x*CHAR_WIDTH,y*CHAR_HEIGHT,cell.ch);
                }else{
                    _draw_character(cr,x*CHAR_WIDTH,y*CHAR_HEIGHT,cell.ch);
                }
                cairo_fill(cr);
            }
        }
    }
    cairo_restore(cr);
}
#endif


#ifdef TEST_RENDER
static void
_tt_render(abydos_plugin_handle_t *h,
           cairo_t *cr,
           int page,
           cairo_bool_t *layer_visibility,
           int varient,
           double time,
           int frame)
{
    _draw_screen(cr,h->cons);
}
#else
static cairo_surface_t *
_tt_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}
#endif


static void
_tt_get_duration(abydos_plugin_handle_t *h,double *frame_duration)
{
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _tt_get_mime_types,
        .new = _tt_new,
        .free = _tt_free,
#if 0
        .progressive_begin = _tt_progressive_begin,
        .progressive_feed = _tt_progressive_feed,
        .progressive_end = _tt_progressive_end,
#else
        .create_from_data = _tt_create_from_data,
#endif
#ifdef TEST_RENDER
        .render = _tt_render,
#else
        .get_image_surface = _tt_get_image_surface,
#endif
        .get_duration = _tt_get_duration
    };
    return &p;
}
