/*
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <endian.h>
#include <stdio.h>
#include <webp/encode.h>
#include <webp/mux.h>

#include "abydos.h"
#include "debug.h"


#if __BYTE_ORDER == __BIG_ENDIAN
static void
_swap_pixels(uint32_t *p,int width,int height,int stride)
{
    int rowskip = stride/sizeof(uint32_t) - width;
    DEBUG_p("row skip: %d",rowskip);
    while(height-->0){
        int x = width;
        while(x-->0){
            *p = le32toh(*p);
            ++p;
        }
        p += rowskip;
    }
}
#endif


static int
_encode_frame(void *src, int w, int h, int stride, WebPData *dst){
#if __BYTE_ORDER == __BIG_ENDIAN
    _swap_pixels(src, w, h, stride);
#endif
    dst->size = WebPEncodeLosslessBGRA(src, w, h, stride, (uint8_t**)&dst->bytes);
    return 0;
}


static int
_encode_animation(abydos_t *ar, int w, int h, WebPData *dst){
    WebPMux* mux;
    WebPMuxError err;
    cairo_surface_t *surface = NULL;
    void *src;
    int stride;
    cairo_t *cr = NULL;
    int i;

    mux = WebPMuxNew();
    if(!mux)
        return -1;

    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,w,h);
    src = cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface);

    cr = cairo_create(surface);
    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
    for(i=0;i<abydos_get_frame_count(ar);++i){
        WebPMuxFrameInfo frame;
        abydos_set_frame(ar,i);
        frame.id = WEBP_CHUNK_ANMF;
        frame.x_offset = frame.y_offset = 0;
        frame.dispose_method = WEBP_MUX_DISPOSE_BACKGROUND;
        frame.blend_method = WEBP_MUX_NO_BLEND;
        frame.duration = abydos_get_delay(ar)*1000;
        abydos_render(ar,cr,0);
        cairo_surface_flush(surface);
        DEBUG_p("encoding frame %d",i);
        _encode_frame(src, w, h, stride, &frame.bitstream);
        err = WebPMuxPushFrame(mux, &frame, 1);
        WebPDataClear(&frame.bitstream);
        if(err != WEBP_MUX_OK){
            DEBUG_p("could not add frame to animation");
            goto out;
        }
    }
    {
        WebPMuxAnimParams anim = {0, 0};
        err = WebPMuxSetAnimationParams(mux, &anim);
        if (err != WEBP_MUX_OK) {
            DEBUG_p("could not set animation parameters");
            goto out;
        }
    }
    err = WebPMuxAssemble(mux, dst);

out:
    WebPMuxDelete(mux);
    if(cr)
        cairo_destroy(cr);
    if(surface)
        cairo_surface_destroy(surface);

    return WEBP_MUX_OK==err ? 0 : -1;
}


#if 1
cairo_status_t
abydos_write_to_webp(abydos_t *ar, const char *filename)
{
    abydos_info_t info;
    WebPData dst;
    int res;

    abydos_get_info(ar,&info);
    if(abydos_get_frame_count(ar)>1){
        res = _encode_animation(ar, info.width, info.height, &dst);
    }else{
        cairo_surface_t *surface = abydos_get_image_surface(ar,0);
        void *src = cairo_image_surface_get_data(surface);
        int stride = cairo_image_surface_get_stride(surface);
        res = _encode_frame(src,info.width,info.height,stride,&dst);
        cairo_surface_destroy(surface);
    }
    if(0==res){
        FILE *f = fopen(filename,"w");
        fwrite(dst.bytes, dst.size, 1, f);
        fclose(f);
    }
    WebPDataClear(&dst);
    return res;
}
#else
cairo_status_t
abydos_write_to_webp(abydos_t *ar, const char *filename)
{
    abydos_info_t info;
    cairo_surface_t *surface = NULL;
    void *data;
    int stride;
    cairo_t *cr;
    int i;

    WebPMux* mux = NULL;
    WebPMuxError err;

    mux = WebPMuxNew();
    if(!mux)
        return -1;

    abydos_get_info(ar,&info);
    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,info.width,info.height);
    data = cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface);

    cr = cairo_create(surface);
    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
    for(i=0;i<abydos_get_frame_count(ar);++i){
        WebPMuxFrameInfo frame;
        abydos_set_frame(ar,i);
        frame.id = WEBP_CHUNK_ANMF;
        frame.x_offset = frame.y_offset = 0;
        frame.dispose_method = WEBP_MUX_DISPOSE_BACKGROUND;
        frame.blend_method = WEBP_MUX_NO_BLEND;
        frame.duration = abydos_get_delay(ar)*1000;
        abydos_render(ar,cr,0);
        cairo_surface_flush(surface);
        DEBUG_p("encoding frame %d",i);
#if __BYTE_ORDER == __BIG_ENDIAN
        _swap_pixels(data, info.width, info.height, stride);
#endif
        frame.bitstream.size = WebPEncodeLosslessBGRA(data, info.width, info.height, stride, (uint8_t**)&frame.bitstream.bytes);
        err = WebPMuxPushFrame(mux, &frame, 1);
        WebPDataClear(&frame.bitstream);
    }
    cairo_destroy(cr);
    {
        WebPMuxAnimParams anim = {0, 0};
        err = WebPMuxSetAnimationParams(mux, &anim);
        if (err != WEBP_MUX_OK) {
            DEBUG_p("could not set animation parameters");
            goto out;
        }
    }
    {
        WebPData webp_data;
        err = WebPMuxAssemble(mux, &webp_data);
        if(WEBP_MUX_OK==err){
            FILE *f = fopen(filename,"w");
            fwrite(webp_data.bytes, webp_data.size, 1, f);
            fclose(f);
        }
        WebPDataClear(&webp_data);
    }


out:
    WebPMuxDelete(mux);
    if(surface)
        cairo_surface_destroy(surface);
    return 0;
}
#endif
