/*
 * Copyright (C) 2018 - 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "../autoconf.h"

#include <dirent.h>
#include <dlfcn.h>                      /* dlopen */
#include <errno.h>
#include <fcntl.h>                      /* open() */
#include <math.h>                       /* fmod() */
#include <pthread.h>
#include <stdio.h>                      /* readdir() */
#include <stdlib.h>                     /* free() */
#include <string.h>                     /* memcpy(), strlen() */
#include <sys/mman.h>                   /* mmap() */
#include <sys/stat.h>
#include <unistd.h>                     /* unlink() */

#include "abydos.h"
#include "abydos-plugin.h"
#include "atomic.h"
#include "debug.h"
#include "hash.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

/**
 * SECTION: abydos
 * @short_description: The basic set of the application API.
 * @title: Basics
 * @include: abydos.h
 *
 * There are three steps for using abydos. The first step is to create an
 * #abydos_t for a specific MIME type using abydos_create(). Then you are
 * done using it you should destroy it using abydos_destroy(). Determining
 * the correct MIME type for your data is outside the scope of abydos. Read
 * more about [image format detection](../detection.html).
 *
 * The second step is to load the data. There are three different methods
 * to choose from for doing this. You can load it from a file using
 * abydos_from_file(). Or you can load it from memory using
 * abydos_from_data(). Or you can use abydos_load_begin(),
 * abydos_load_feed() and abydos_load_end() to progressively load the image.
 * With the last option you may get the opertunity to display a partially
 * decoded image. But this is not always possible since some images store
 * crucial information at the end of the file.
 *
 * The third step is to get the image data. There are two different ways to
 * do this. Either you can render it to a #cairo_t using abydos_render().
 * This is the recommended approch if you intend to display the image since
 * vector graphics will always be rendered in full quality intependant of
 * scale. Or you can get a cairo image surface with
 * abydos_get_image_surface().
 * |[
#include <abydos.h>
#include <cairo.h>

int main (int argc, char **argv)
{
    abydos_t *ar;
    cairo_surface_t *surface;

    ar = abydos_create ("image/x-ilbm");
    if (!ar)
        return 1;
    if (abydos_from_file (ar, "example.ilbm") < 0)
        return 1;
    surface = abydos_get_image_surface (ar);
    abydos_destroy (ar);
    cairo_surface_write_to_png (surface, "example.png");
    cairo_surface_destroy (surface);
    return 0;
}
 * ]|
 **/


typedef struct{
    AtomicInt refcount;
    abydos_plugin_interface_t* vt;
    abydos_plugin_handle_t *handle;
    abydos_plugin_info_t info;
    /* progressive loading */
    struct{
        abydos_info_callback_t info_func;
        abydos_update_callback_t update_func;
        void *userdata;
        FILE *f;
        union{
            char *buffer;
            char *tmpfile;
        };
        size_t buffer_len;
    }load;
    struct{
        abydos_size_t *variant_size;
        int default_variant;
        /* animation */
        double *frame_duration;
        double *frame_time;
        double total_duration;
    }done;
} immutable_t;

struct _abydos_t{
    /*< private >*/
    AtomicInt refcount;
    immutable_t *im;
    cairo_surface_t* cached_surface;
    int page;
    cairo_bool_t *layer_visibility ;
    int prefered_variant;
    double time;
    int frame;
};

typedef struct{
    void *handle;
    abydos_plugin_interface_t *vt;
} plugin_t;


static HashMap *plugin_map;


static void
_free_plugin(plugin_t *p)
{
    if(p->vt->done)
        p->vt->done();
    dlclose(p->handle);
    free(p);
}


static void
_try_plugin(const char *filename)
{
    void *handle;
    abydos_plugin_func_t func;
    plugin_t *plugin;
    const char **mime_types;
    int i;
    DEBUG_p("loading %s",filename);
    handle = dlopen(filename,RTLD_NOW);
    if(!handle){
        DEBUG_p("dlopen failed: %s",dlerror());
        return;
    }
    func = dlsym(handle,"Abydos0");
    if(!func){
        dlclose(handle);
        return;
    }
    plugin = (plugin_t*)malloc(sizeof(plugin_t));
    plugin->handle = handle;
    plugin->vt = func();
    /* check if all required functions are implemented */
    if(  !plugin->vt->get_mime_types ||
         !plugin->vt->free ||
       ((!plugin->vt->progressive_begin ||
         !plugin->vt->progressive_feed ||
         !plugin->vt->progressive_end) &&
         !plugin->vt->create_from_data &&
         !plugin->vt->create_from_file) ||
        (!plugin->vt->render &&
         !plugin->vt->get_image_surface))
    {
        DEBUG_p("not all required functions are provided");
        free(plugin);
        dlclose(handle);
        return;
    }
    mime_types = plugin->vt->get_mime_types();
    for(i=0;mime_types[i];++i){
        if(hash_map_get(plugin_map,mime_types[i])){
            DEBUG_p("already have plugin for %s",mime_types[i]);
            continue;
        }
        DEBUG_p("register %s",mime_types[i]);
        hash_map_add(plugin_map,(void*)mime_types[i],plugin);
    }
}


static cairo_bool_t
_open_tmpfile(char **r_filename,FILE** r_f)
{
    static const char template[] = PACKAGE "-XXXXXX";
    const char *tmpdir;
    size_t len;
    char *filename;
    int fd;
    tmpdir = getenv("XDG_RUNTIME_DIR");
    if(!tmpdir)
        tmpdir = getenv("TMPDIR");
    if(!tmpdir)
        tmpdir = "/tmp";
    len = strlen(tmpdir);
    filename = (char*)malloc(len+1+sizeof(template));
    memcpy(filename,tmpdir,len);
    filename[len] = '/';
    memcpy(filename+len+1,template,sizeof(template));
    fd = mkstemp(filename);
    if(fd<0){
        free(filename);
        return 0;
    }
    *r_filename = filename;
    *r_f = fdopen(fd,"w+");
    return 1;
}


static void
_get_size(immutable_t *im, double *width, double *height, int flags)
{
    *width = im->info.width;
    *height = im->info.height;
    switch(flags & ABYDOS_ASPECT_RATIO_MASK){
    case ABYDOS_IGNORE_ASPECT_RATIO:
        break;
    case ABYDOS_SHRINK_TO_ASPECT_RATIO:
        if(im->info.pixel_ratio < 1){
            /* high pixels */
            *width *= im->info.pixel_ratio;
        }else if(im->info.pixel_ratio > 1){
            /* wide pixels */
            *height /= im->info.pixel_ratio;
        }
        break;
    case ABYDOS_GROW_TO_ASPECT_RATIO:
    default:
        if(im->info.pixel_ratio < 1){
            /* high pixels */
            *height /= im->info.pixel_ratio;
        }else if(im->info.pixel_ratio > 1){
            /* wide pixels */
            *width *= im->info.pixel_ratio;
        }
        break;
    }
}


static void
_get_size_int(immutable_t *im, int *r_width, int *r_height, int flags)
{
    double width;
    double height;
    _get_size(im,&width,&height,flags);
    if(r_width)
        *r_width = width+0.5;
    if(r_height)
        *r_height = height+0.5;
}



static void
abydos_init()
{
    DIR *d;
    struct dirent *e;
    size_t len1;
    plugin_map = hash_map_new(str_hash,str_equal,NULL,(HashMapNotifyFunc)_free_plugin);

    DEBUG_p("searching for plugins in %s",PLUGINDIR);
    d = opendir(PLUGINDIR);
    if(!d)
        return;
    len1 = strlen(PLUGINDIR);
    while ((e = readdir(d))) {
        size_t len2;
        char *path;
        if('.' == e->d_name[0])
            continue;
        len2 = strlen(e->d_name);
        path = (char*)malloc(len1 + 1 + len2 + 1);
        memcpy(path,PLUGINDIR,len1);
        path[len1] = '/';
        memcpy(path+len1+1,e->d_name,len2);
        path[len1+1+len2] = '\0';
        _try_plugin(path);
        free(path);
    }
    closedir(d);
}


static immutable_t *
immutable_create(const char *mime_type)
{
    immutable_t *im;
    plugin_t *p = (plugin_t*)hash_map_get(plugin_map,mime_type);
    if(!p){
        DEBUG_p("no loader found for mime type '%s'",mime_type);
        return NULL;
    }
    im = (immutable_t*)malloc(sizeof(immutable_t));
    im->info.version = ABYDOS_PLUGIN_INFO_VERSION;
    im->info.error = NULL;

    /* setup default values */
    im->info.width = 0;
    im->info.height = 0;
    im->info.pixel_ratio = 1;
    im->info.page_count = 1;
    im->info.layer_count = 1;
    im->info.variant_count = 1;
    im->info.frame_count = 1;
    im->info.scalable_size = 0;
    im->info.scalable_time = 0;
    im->info.threadsafe = 0;

    im->handle = p->vt->new(mime_type,&im->info);
    if(!im->handle){
        free(im);
        return NULL;
    }

    im->done.variant_size = NULL;
    im->done.frame_duration = NULL;
    im->done.frame_time = NULL;

    im->refcount = 1;
    im->vt = p->vt;
    return im;
}


/**
 * abydos_create:
 * @mime_type: MIME type
 *
 * Creates a new #abydos_t object to load an image of the type specified by @mime_type.
 *
 * returns: An abydos_t capable of handling @mime_type. Or NULL if none is
 *          found. Call abydos_destroy() then you are done using it.
 **/
abydos_t *
abydos_create(const char *mime_type)
{
    static int initialized;
    static pthread_mutex_t init_lock = PTHREAD_MUTEX_INITIALIZER;
    immutable_t *im;
    abydos_t *ar;
    if(!initialized){
        pthread_mutex_lock(&init_lock);
        if(!initialized){
            abydos_init();
            initialized = 1;
        }
        pthread_mutex_unlock(&init_lock);
    }
    im = immutable_create(mime_type);
    if(!im)
        return NULL;
    ar = (abydos_t*)malloc(sizeof(abydos_t));
    ar->refcount = 1;
    ar->im = im;
    ar->cached_surface = NULL;
    ar->page = 0;
    ar->prefered_variant = -1;
    ar->time = 0;
    ar->frame = 0;

    return ar;
}


/**
 * abydos_clone
 * @src: An #abydos_t to clone
 *
 * Make a clone of @src, which is cheap and effectively only copies the
 * internal state. This is useful for maintaining multiple states at once.
 * Such as rendering several pages of a multi page image in parallell.
 *
 * returns: A clone of @src. Call abydos_destroy() then you are done using
            it.
 **/
abydos_t *
abydos_clone(abydos_t *src)
{
    abydos_t *dst = (abydos_t*)malloc(sizeof(abydos_t));
    memcpy(dst,src,sizeof(abydos_t));
    atomic_int_inc(&dst->im->refcount);
    if(dst->layer_visibility){
        dst->layer_visibility = (cairo_bool_t*)malloc(dst->im->info.layer_count*sizeof(cairo_bool_t));
        memcpy(dst->layer_visibility,src->layer_visibility,dst->im->info.layer_count*sizeof(cairo_bool_t));
    }
    return dst;
}


/**
 * abydos_reference
 * @ar: an #abydos_t
 *
 * Increases the reference count on @ar by one. This prevents ar from being
 * destroyed until a matching call to abydos_destroy() is made.
 *
 * returns: The same #abydos_t. Call abydos_destroy() then you are done
            using it.
 **/
abydos_t *
abydos_reference(abydos_t *ar)
{
    atomic_int_inc(&ar->refcount);
    return ar;
}


/**
 * abydos_destroy
 * @ar: an #abydos_t
 *
 * Decreases the reference count on @ar by one. If the result is zero, then
 * @ar and all associated resources are freed.
 */
void
abydos_destroy(abydos_t *ar)
{
    if(atomic_int_dec_and_test(&ar->refcount)){
        if(atomic_int_dec_and_test(&ar->im->refcount)){
            ar->im->vt->free(ar->im->handle);
            if(ar->im->done.variant_size)
                free(ar->im->done.variant_size);
            if(ar->im->done.frame_duration){
                free(ar->im->done.frame_duration);
                free(ar->im->done.frame_time);
            }
            free(ar->im);
        }
        free(ar);
    }
}

/* --- loading ---------------------------------------------------------- */

static void
_setup_immutable(immutable_t *im)
{
    int i;
    /* setup variants */
    if(im->info.variant_count > 1){
        int best_value = -1;
        im->done.variant_size = (abydos_size_t*)malloc(im->info.variant_count*sizeof(abydos_size_t));
        if(im->vt->get_variant_size){
            im->vt->get_variant_size(im->handle,im->done.variant_size);
            for(i=0;i<im->info.variant_count;++i){
                int value = im->done.variant_size[i].width * im->done.variant_size[i].height;
                if(value > best_value){
                    im->done.default_variant = i;
                    best_value = value;
                }
                if(im->done.variant_size[i].width > im->info.width)
                    im->info.width = im->done.variant_size[i].width;
                if(im->done.variant_size[i].height > im->info.height)
                    im->info.height = im->done.variant_size[i].height;
            }
        }else{
            for(i=0;i<im->info.variant_count;++i){
                im->done.variant_size[i].width = im->info.width;
                im->done.variant_size[i].height = im->info.height;
            }
            im->done.default_variant = 0;
        }
    }else{
        im->done.variant_size = NULL;
        im->done.default_variant = 0;
    }
    /* setup animation */
    im->done.total_duration = 0;
    if(im->info.frame_count > 1){
        im->done.frame_duration = (double*)malloc(im->info.frame_count*sizeof(double));
        im->done.frame_time = (double*)malloc(im->info.frame_count*sizeof(double));
        if(im->vt->get_duration){
            im->vt->get_duration(im->handle,im->done.frame_duration);
            for(i=0;i<im->info.frame_count;++i){
                if(im->done.frame_duration[i] < 0.02)
                    im->done.frame_duration[i] = 0.02;
                im->done.frame_time[i] = im->done.total_duration;
                im->done.total_duration += im->done.frame_duration[i];
            }
        }else{
            for(i=0;i<im->info.frame_count;++i){
                im->done.frame_duration[i] = 50;
                im->done.frame_time[i] = 50 * i;
            }
            im->done.total_duration = 50 * im->info.frame_count;
        }
    }else{
        im->done.frame_duration = NULL;
        im->done.frame_time = NULL;
    }
}


static void
_setup(abydos_t *ar)
{
    _setup_immutable(ar->im);
    if(ar->im->info.layer_count > 1){
        ar->layer_visibility = (cairo_bool_t*)malloc(ar->im->info.layer_count * sizeof(cairo_bool_t));
        if(ar->im->vt->get_layer_visibility){
            ar->im->vt->get_layer_visibility(ar->im->handle,ar->layer_visibility);
        }else{
            int i;
            for(i=0;i<ar->im->info.layer_count;++i)
                ar->layer_visibility[i] = 1;
        }
    }else{
        ar->layer_visibility = NULL;
    }
#if 0
    /* setup layers */
    if(im->info.layer_count > 1){
        ar->layer_visibility = (cairo_bool_t*)malloc(im->info.layer_count*sizeof(cairo_bool_t));
        /*
        for(i=0;i<im->info.layer_count;++i)
            ar->layer_visibility[i] = ar->layer_visibility[i];
        */
    }else{
        ar->layer_visibility = NULL;
    }
#endif
}


/**
 * abydos_from_data
 * @ar: an #abydos_t
 * @data: the data containing the image
 * @len: length of the data
 *
 * Create the image from a single chunk of data. If you prefer to feed the
 * data a little at a time, use progressive interface abydos_load_begin(),
 * abydos_load_feed() and abydos_load_end(). If the operation fails you can
 * call abydos_error() to get an error message. Apart fram that nothing more
 * can be done with @ar and it must be destroyed.
 *
 * returns: 0 on success. Otherwise the operation has failed.
 */
int
abydos_from_data(abydos_t *ar, const char *data, size_t len)
{
    immutable_t *im = ar->im;
    if(im->vt->create_from_data){
        if(im->vt->create_from_data(im->handle,data,len)<0)
            return -1;
    }else if(im->vt->progressive_begin && im->vt->progressive_feed && im->vt->progressive_end){
        im->vt->progressive_begin(im->handle,NULL,NULL,NULL);
        if(im->vt->progressive_feed(im->handle,data,len) < 0)
            return -1;

        if(im->vt->progressive_end(im->handle)<0)
            return -1;
    }else{
        FILE *f;
        char *tmpfile;
        int res;
        _open_tmpfile(&tmpfile,&f);
        res = fwrite(data,1,len,im->load.f) < len ? -1 : 0;
        fclose(f);
        if(res >= 0)
            res = abydos_from_file(ar,tmpfile);
        unlink(tmpfile);
        free(tmpfile);
        if(res < 0)
            return res;
    }
    _setup(ar);
    return 0;
}


/**
 * abydos_from_file
 * @ar: an #abydos_t
 *
 * Create the image from a file. You may want to use the convenience
 * function abydos_create_from_file(), which is identical to calling
 * abydos_create() followed by abydos_from_file(). If the operation fails
 * you can call abydos_error() to get an error message. Apart fram that
 * nothing more can be done with @ar and it must be destroyed.
 *
 * returns: 0 on success. Otherwise the operation has failed.
 */
int
abydos_from_file(abydos_t *ar, const char *filename)
{
    immutable_t *im = ar->im;
    struct stat st;
    int fd;
    void *data;
    int res;

    if(im->vt->create_from_file){
        if(im->vt->create_from_file(im->handle,filename)<0)
            return -1;
        _setup(ar);
        return 0;
    }

    if(stat(filename,&st)<0){
        im->info.error = strerror(errno);
        return -1;
    }
    fd = open(filename,O_RDONLY);
    if(fd<0){
        im->info.error = strerror(errno);
        return -1;
    }
    data = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE | MAP_POPULATE, fd, 0);
    if(MAP_FAILED == data){
        close(fd);
        im->info.error = strerror(errno);
        return -1;
    }
    res = abydos_from_data(ar, (const char*)data, st.st_size);
    munmap(data,st.st_size);
    close(fd);
    return res;
}


static void
on_info(void *userdata)
{
    abydos_t *ar = (abydos_t*)userdata;
    _setup(ar);
    if(ar->im->load.info_func)
        ar->im->load.info_func(ar->im->load.userdata,ar);
}


static void
on_update(void *userdata, cairo_rectangle_int_t *r)
{
    abydos_t *ar = (abydos_t*)userdata;
    if(ar->im->load.update_func)
        ar->im->load.update_func(ar->im->load.userdata,ar,r);
}


/**
 * abydos_load_begin
 * @ar: an #abydos_t
 * @info_func: a function to be called as soon as abydos_get_info() can be
 *             called, or NULL
 * @update_func: a function to be called every time the image has been
 *               updated, or NULL
 * @userdata: user data to be passed to @info_func and @update_func
 *
 * Begin progressive loading of an image. If you pass @info_func and
 * @update_func they are guaranteed to be called before abydos_load_end()
 * returns successfully. @info_func() will be called first and exactly once.
 * @update_func() will be called at least once. If abydos_load_feed() or
 * abydos_load_end() fails @info_func() and @update_func() may or may not be
 * called.
 **/
void
abydos_load_begin(abydos_t *ar,abydos_info_callback_t info_func,abydos_update_callback_t update_func,void *userdata)
{
    immutable_t *im = ar->im;
    im->load.info_func = info_func;
    im->load.update_func = update_func;
    im->load.userdata = userdata;
    if(im->vt->progressive_begin && im->vt->progressive_feed && im->vt->progressive_end){
        im->vt->progressive_begin(im->handle,on_info,on_update,ar);
    }else if(im->vt->create_from_data){
        im->load.f = open_memstream(&im->load.buffer,&im->load.buffer_len);
    }else{
        _open_tmpfile(&im->load.tmpfile,&im->load.f);
    }
}


/**
 * abydos_load_feed
 * @ar: an #abydos_t
 * @data: the data containing a part of the image
 * @len: length of the data
 *
 * Feed bytes of the image. If the operation fails you can call
 * abydos_error() to get an error message. Apart fram that nothing more can
 * be done with @ar and it must be destroyed.
 *
 * returns: 0 on success. Otherwise the operation has failed.
 */
int
abydos_load_feed(abydos_t *ar, const char* data, size_t len)
{
    immutable_t *im = ar->im;
    if(im->info.error)
        return -1;

    if(im->vt->progressive_begin && im->vt->progressive_feed && im->vt->progressive_end){
        return im->vt->progressive_feed(im->handle,data,len);
    }else if(im->vt->create_from_data || im->vt->create_from_file){
        return fwrite(data,1,len,im->load.f) < len ? -1 : 0;
    }
    return -1;
}


/**
 * abydos_load_end
 * @ar: an #abydos_t
 *
 * End progressive loading of an image. If the operation fails you can call
 * abydos_error() to get an error message. Apart fram that nothing more can
 * be done with @ar and it must be destroyed.
 *
 * returns: 0 on success. Otherwise the operation has failed.
 **/
int
abydos_load_end(abydos_t *ar)
{
    immutable_t *im = ar->im;
    if(im->info.error)
        return -1;

    if(im->vt->progressive_begin && im->vt->progressive_feed && im->vt->progressive_end){
        if(im->vt->progressive_end(im->handle)<0)
            return -1;
        _setup(ar);
    }else{
        int res;
        fclose(im->load.f);
        if(im->vt->create_from_data){
            res = abydos_from_data(ar,im->load.buffer,im->load.buffer_len);
            free(im->load.buffer);
        }else{
            res = abydos_from_file(ar,im->load.tmpfile);
            unlink(im->load.tmpfile);
            free(im->load.tmpfile);
        }
        if(res < 0)
            return res;
        if(im->load.info_func)
            im->load.info_func(im->load.userdata,ar);
        if(im->load.update_func){
            cairo_rectangle_int_t rect;
            rect.x = 0;
            rect.y = 0;
            /* FIXME!! flags from the user is needed here. */
            _get_size_int(im,&rect.width,&rect.height,0);
            im->load.update_func(im->load.userdata,ar,&rect);
        }
    }
    return 0;
}

/* --- info ------------------------------------------------------------- */

void
abydos_get_info_V(abydos_t *ar,abydos_info_t *info,int info_version)
{
    immutable_t *im = ar->im;
    /* FIXME!! a flags argument is needed here */
    _get_size_int(im,&info->width,&info->height,0);
    info->features = (1==im->info.pixel_ratio?ABYDOS_FEATURE_SQUARE_PIXELS:0)
                    |(im->info.scalable_size?ABYDOS_FEATURE_SCALABLE_SIZE:0);
    info->pixel_ratio = im->info.pixel_ratio;
}


/**
 * abydos_get_size
 * @ar: an #abydos_t
 *
 * Get the size of a loaded image. The @flags may affect the @width and
 * @height if the image has non square pixels. Valid flags are one of:
 * - %ABYDOS_IGNORE_ASPECT_RATIO
 * - %ABYDOS_SHRINK_TO_ASPECT_RATIO
 * - %ABYDOS_GROW_TO_ASPECT_RATIO
 *
 * The default is to grow to aspect ratio, but it may change in future
 * versions.
 */
void
abydos_get_size(abydos_t *ar,int *r_width,int *r_height,int flags)
{
    _get_size_int(ar->im,r_width,r_height,flags);
}


const char *
abydos_error(abydos_t *ar)
{
    return ar->im->info.error;
}

/* --- drawing ---------------------------------------------------------- */

static void
abydos_render_internal(immutable_t *im,cairo_t *cr,int page,cairo_bool_t *visible_layer,int variant,double time,int frame)
{
    if(variant<0 && im->info.variant_count>1){
        cairo_matrix_t matrix;
        double scale;
        int best_variant = im->done.default_variant;
        double best_scale = 1;
        int i;
        cairo_get_matrix(cr,&matrix);
        scale = MAX(sqrt(matrix.xx*matrix.xx+matrix.yx*matrix.yx),
                    sqrt(matrix.xy*matrix.xy+matrix.yy*matrix.yy));
        for(i=0;i<im->info.variant_count;++i){
            double this_scale_x = (double)im->done.variant_size[i].width/im->info.width;
            double this_scale_y = (double)im->done.variant_size[i].height/im->info.height;
            double this_scale = MIN(this_scale_x,this_scale_y);
            if(this_scale >= scale && this_scale < best_scale){
                best_variant = i;
                best_scale = this_scale;
            }
        }
        variant = best_variant;
        cairo_scale(cr,1/best_scale,1/best_scale);
    }
    if(im->vt->render){
        im->vt->render(im->handle,cr,page,visible_layer,variant,time,frame);
    }else{
        cairo_surface_t *src = im->vt->get_image_surface(im->handle,page,visible_layer,variant,time,frame);
        cairo_pattern_t *pattern = cairo_pattern_create_for_surface(src);
        cairo_set_source(cr, pattern);
        cairo_pattern_set_filter(pattern, CAIRO_FILTER_BEST);
        cairo_paint(cr);
        cairo_pattern_destroy(pattern);
        cairo_surface_destroy(src);
    }
}


/**
 * abydos_render
 * @ar: an #abydos_t
 *
 * Render the image to a #cairo_t. Use ordinary cairo trasformations to
 * scale the image. The @flags may affect the size of the image if it has
 * non square pixels. Valid flags are one of:
 * - #ABYDOS_IGNORE_ASPECT_RATIO
 * - #ABYDOS_SHRINK_TO_ASPECT_RATIO
 * - #ABYDOS_GROW_TO_ASPECT_RATIO
 *
 * The default is to grow to aspect ratio, but it mat change in future
 * versions.
 */
void
abydos_render(abydos_t *ar,cairo_t *cr,int flags)
{
    immutable_t *im = ar->im;
    double width;
    double height;
    _get_size(im,&width,&height,flags);
    cairo_save(cr);
    if(width != im->info.width || height != im->info.height)
        cairo_scale(cr,width/im->info.width,height/im->info.height);
    abydos_render_internal(ar->im,cr,ar->page,ar->layer_visibility,ar->prefered_variant,ar->time,ar->frame);
    cairo_restore(cr);
}


static cairo_surface_t *
abydos_get_image_surface_internal(immutable_t *im,int flags,int page,cairo_bool_t *visible_layer,int variant,double time,int frame)
{
    cairo_t *cr;
    double width;
    double height;
    /* Always fall back to the default variant. */
    if(variant<0)
        variant = im->done.default_variant;
    _get_size(im,&width,&height,flags);
    if(width==im->info.width && height==im->info.height)
        if(im->vt->get_image_surface)
            return im->vt->get_image_surface(im->handle,page,visible_layer,variant,time,frame);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    cr = cairo_create(surface);
    if(width != im->info.width || height != im->info.height)
        cairo_scale(cr,width/im->info.width,height/im->info.height);
    abydos_render_internal(im,cr,page,visible_layer,variant,time,frame);
    cairo_destroy(cr);
    return surface;
}


/**
 * abydos_get_image_surface
 * @ar: an #abydos_t
 *
 * Get a #cairo_surface_t containing the image. This may or may not be a
 * reference to an internally cached image surface. If you don't intend to
 * modify it or if you intend to get it and modify it only once, then this
 * is no problem. But you intend to get several copies and modify them you
 * need to pass one of the flags %ABYDOS_PREFER_COPY or
 * %ABYDOS_PREFER_TRANSFER. The @flags may also affect the size of the image
 * if it has non square pixels. Valid flags are one of:
 * - %ABYDOS_PREFER_SHARE
 * - %ABYDOS_PREFER_COPY
 * - %ABYDOS_PREFER_TRANSFER
 *
 * and one of:
 * - %ABYDOS_IGNORE_ASPECT_RATIO
 * - %ABYDOS_SHRINK_TO_ASPECT_RATIO
 * - %ABYDOS_GROW_TO_ASPECT_RATIO
 *
 * The default is %ABYDOS_PREFER_SHARE and %ABYDOS_GROW_TO_ASPECT_RATIO, but
 * it may change in the future.
 *
 * For the common simple cases you might want to use the convinience
 * function abydos_load() to quickly get an image surface from a file.
 *
 * returns: If the image loading has suceeded a valid @cairo_surface_t will
 *          be returned. Call cairo_surface_destroy() then you are done
 *          using it. If the image loading hasn't succeeded the return value
 *          is undefined.
 */
cairo_surface_t *
abydos_get_image_surface(abydos_t *ar,int flags)
{
    cairo_surface_t *surface;
    if(ar->cached_surface)
        return cairo_surface_reference(ar->cached_surface);
    surface = abydos_get_image_surface_internal(ar->im,flags,ar->page,ar->layer_visibility,ar->prefered_variant,ar->time,ar->frame);
    if(!ar->im->vt->get_image_surface)
        ar->cached_surface = cairo_surface_reference(surface);
    return surface;
}

/* --- multi page --- */
/**
 * SECTION: abydos-pages
 * @short_description: Functions for handling images with multiple pages.
 * @title: Pages
 * @include: abydos.h
 *
 * Some image formats can caontain multiple pages. The definition of pages
 * in this context is if the image contains more than one sub image, which
 * are not frames of an animation, layers of the same image nor the same
 * image in different sizes. It could to subimages which are defined as
 * pages by the format (such as in the case of PDF documents) but it could
 * also be formats that simply stores severaral images together in a file in
 * a more loosely defined manner. Some formats may have an internal
 * organization of the sub images in the form of a tree. But abydos will in
 * that case always fletten the tree and index the pages sequentially.
 **/

/**
 * abydos_get_page_count
 * @ar: an #abydos_t
 *
 * Get the number of pages of an image.
 **/
int
abydos_get_page_count(abydos_t *ar)
{
    return ar->im->info.page_count;
}


/**
 * abydos_get_page
 * @ar: an #abydos_t
 *
 * Returns the index of the currently selected page.
 **/
int
abydos_get_page(abydos_t *ar)
{
    return ar->page;
}


/**
 * abydos_set_page
 * @ar: an #abydos_t
 *
 * Selects a page.
 */
void
abydos_set_page(abydos_t *ar,int page)
{
    if(page<0){
        page = 0;
    }else if(page >= ar->im->info.page_count){
        page = ar->im->info.page_count - 1;
    }
    ar->page = page;
}

/* --- layers ----------------------------------------------------------- */
/**
 * SECTION: abydos-layers
 * @short_description: Functions for handling image layers.
 * @title: Layers
 * @include: abydos.h
 *
 * Some formats may compose an image out of several layers. It could simply
 * be a set of partly transparent pixmaps placed on top of each other. But
 * layers could also be more like filter operations. Sometimes the internal
 * organization of the layers may be in the form of a tree. But abydos will
 * in that case always flatten the tree and index the layers sequentially.
 **/

/**
 * abydos_get_layer_count
 * @ar: an #abydos_t
 *
 * Get the number of layers of the image.
 */
int
abydos_get_layer_count(abydos_t *ar)
{
    return ar->im->info.layer_count;
}


int abydos_get_layer_visibility(abydos_t *ar,int layer)
{
    if(ar->layer_visibility)
        return ar->layer_visibility[layer];
    return 1;
}

/* --- variants --------------------------------------------------------- */
/**
 * SECTION: abydos-variants
 * @short_description: Functions for handling different size variants.
 * @title: Variants
 * @include: abydos.h
 *
 * Some formats may contain different varaints of the same image. The
 * definition of variant in this context is different versions of the same
 * image in different sizes. It could be embedded thumbnails. The thumbnail
 * might not just be a scaled down version of the full size image, it could
 * be some intelligent cropping involved. Or it could be a set of icons with
 * different sizes and different detail level. Or it could be mipmap
 * textures.
 **/

/**
 * abydos_get_variant_count
 * @ar: an #abydos_t
 *
 * Get the number of variants of the image.
 **/
int
abydos_get_variant_count(abydos_t *ar)
{
    return ar->im->info.variant_count;
}


void
abydos_get_variant_size(abydos_t *ar,int variant,int *width, int *height)
{
    if(ar->im->info.variant_count > 1){
        variant %= ar->im->info.variant_count;
        if(width)
            *width = ar->im->done.variant_size[variant].width;
        if(height)
            *height = ar->im->done.variant_size[variant].height;
    }else{
        if(width)
            *width = ar->im->info.width;
        if(height)
            *height = ar->im->info.height;
    }
}


int abydos_get_variant_at_size_min(abydos_t *ar,int width, int height)
{
    /* FIXME!! */
    return 0;
}


int
abydos_get_variant_at_size_max(abydos_t *ar,int width, int height)
{
    /* FIXME!! */
    return 0;
}


void
abydos_set_variant(abydos_t *ar,int variant)
{
    ar->prefered_variant = variant>=0 ? variant%ar->im->info.variant_count : -1;
}


void
abydos_prefer_variant(abydos_t *ar,int width,int height)
{
    /* FIXME!! */
}

/* --- animation -------------------------------------------------------- */
/**
 * SECTION: abydos-animation
 * @short_description: Functions for handling animation.
 * @title: Animation
 * @include: abydos.h
 *
 * Two approaches can be used for handling animations. Either by frame or by
 * time. The latter is usually more convinient for displaying the animation.
 * Abydos can tell you for how long you should display a frame, or how long
 * you should wait before updating the time of the animation and redraw it.
 * But measuring time is beyond the scope of abydos. Basically you do
 * something like this.
 *|[
    last_time = now();
    for(;;){
        abydos_render(ar,cr,0);
        delay(abydos_get_delay(ar));
        this_time = now();
        abydos_add_time(ar,this_time - last_time)
        last_time = this_time;
    }
 *]|
 * Of course you don't want to use a simple delay in a real application.
 * Instead you should use some kind of event scheduler to get a callback.
 * Please also note that if you need to round the delay time you get from
 * abydos_get_delay() you must round it up. Otherwise the time of the
 * animation will be set to right before the frame change (and the next time
 * it will be rounded down to zero and you will get nowere).
 **/

/**
 * abydos_get_frame_count
 * @ar: an #abydos_t
 *
 * Get the number of frames of an animation. Some animations may not be
 * based around the concept of frames but are continous to their nature. In
 * this case %ABYDOS_FEATURE_SCALABLE_TIME will be set and frames will be
 * placed at reasonable intervals. It is also possible that
 * %ABYDOS_FEATURE_SCALABLE_TIME is set for singe frame images (and duration
 * of zero). In that case the image will still be animated somehow, but
 * depending on some external variable (such as the clock time).
 **/
int
abydos_get_frame_count(abydos_t *ar)
{
    return ar->im->info.frame_count;
}


/**
 * abydos_get_total_duration
 * @ar: an #abydos_t
 *
 * Get the total duration of an animation.
 *
 * returns: Total duration in seconds.
 */
double
abydos_get_total_duration(abydos_t *ar)
{
    return ar->im->done.total_duration;
}


/**
 * abydos_get_frame_at_time
 * @ar: an #abydos_t
 * @sec: time in seconds
 *
 * Get the index if the frame at @sec. Note that this cannot simply be
 * calculated by dividing @sec with the duration of the first frame, since
 * frames may have different duration.
 **/
int
abydos_get_frame_at_time(abydos_t *ar,double sec)
{
    immutable_t *im = ar->im;
    int frame;
    sec = fmod(sec,im->done.total_duration);
    for(frame=0;im->done.frame_time[frame]+im->done.frame_duration[frame] <= sec && frame < im->info.frame_count;++frame);
    return frame;
}


/**
 * abydos_get_time_for_frame
 * @ar: an #abydos_t
 *
 * Get the time (into the animation) at which @frame should appear.
 *
 * returns: time in seconds
 **/
double
abydos_get_time_for_frame(abydos_t *ar,int frame)
{
    return ar->im->done.frame_time[frame];
}


/**
 * abydos_get_frame
 * @ar: an #abydos_t
 *
 * Returns the index of the current frame.
 **/
int
abydos_get_frame(abydos_t *ar)
{
    return ar->frame;
}


/**
 * abydos_get_time
 * @ar: an #abydos_t
 *
 * Get the current time into the animation.
 *
 * @returns: The time in seconds.
 **/
double
abydos_get_time(abydos_t *ar)
{
    return ar->time;
}


/**
 * abydos_set_frame
 * @ar: an #abydos_t
 *
 * Select @frame which will also implicitly set the time of the animation to
 * the time for this frame.
 **/
void
abydos_set_frame(abydos_t *ar,int frame)
{
    ar->time = ar->im->done.frame_time[frame];
    ar->frame = frame;
}


/**
 * abydos_set_time
 * @ar: an #abydos_t
 *
 * Set the time of the animation in seconds. It will also implicitly set the
 * corresponding frame.
 *
 * @returns: True if the the frame has changed
 **/
cairo_bool_t
abydos_set_time(abydos_t *ar,double sec)
{
    int new_frame;
    cairo_bool_t res;
    /* Add half a microsecond to compensate for rounding errors causing the
     * call to be made just a little bit too early. */
    sec += 0.0000005;
    DEBUG_p("requested time %f",sec);
    sec = fmod(sec,ar->im->done.total_duration);
    DEBUG_p("current time %f, total duration = %f",sec,ar->im->done.total_duration);
    new_frame = abydos_get_frame_at_time(ar,sec);
    DEBUG_p("old frame %d, new frame %d",ar->frame,new_frame);
    ar->time = sec;
    res = ar->frame != new_frame || ar->im->info.scalable_time;
    ar->frame = new_frame;
    return res;
}


/**
 * abydos_next_frame
 * @ar: an #abydos_t
 *
 * Advance the animation one frame and increse the time accordingly.
 **/
void
abydos_next_frame(abydos_t *ar)
{
    if(++ar->frame >= ar->im->info.frame_count)
        ar->frame = 0;
    ar->time = ar->im->done.frame_time[ar->frame];
}


/**
 * abydos_add_time
 * @ar: an #abydos_t
 *
 * Advance the animation @sec seconds. It will also implicitly set the
 * corresponding frame.
 *
 * @returns: True if the the frame has changed
 */
cairo_bool_t
abydos_add_time(abydos_t *ar,double sec)
{
    DEBUG_p("adding %f seconds",sec);
    return abydos_set_time(ar,ar->time+sec);
}


/**
 * abydos_get_delay
 * @ar: an #abydos_t
 *
 * Get the amount of time until the next frame in seconds. Note that if this
 * value needs to be rounded for some reason, it must be rounded up
 * (otherwise the next time advancement will land right before the frame
 * change).
 **/
double
abydos_get_delay(abydos_t *ar)
{
    double sec;
    double delay;
    if(ar->im->info.frame_count <= 1)
        return -1;
    sec = ar->time - ar->im->done.frame_time[ar->frame];
    DEBUG_p("current time : %f",ar->time);
    DEBUG_p("current_frame: %d",ar->frame);
    DEBUG_p("current_frame time: %f",ar->im->done.frame_time[ar->frame]);
    DEBUG_p("current_frame duration: %f",ar->im->done.frame_duration[ar->frame]);
    DEBUG_p("%f sec into current frame (%d)",sec,ar->frame);
    delay = ar->im->done.frame_duration[ar->frame] - sec;
    DEBUG_p("%f sec left of frame duration",delay);
    return delay;
}
